#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""This module provides functions to interface with scipy.sparse."""
import itertools
from functools import reduce
import numpy.linalg
import numpy

import scipy
import scipy.sparse
import scipy.sparse.linalg

from openfermion.ops.operators import (FermionOperator, QubitOperator,
                                       BosonOperator, QuadOperator)
from openfermion.ops.representations import (DiagonalCoulombHamiltonian,
                                             PolynomialTensor)
from openfermion.transforms.opconversions import normal_ordered
from openfermion.utils.indexing import up_index, down_index
from openfermion.utils.operator_utils import count_qubits, is_hermitian

# Make global definitions.
identity_csc = scipy.sparse.identity(2, format='csc', dtype=complex)
pauli_x_csc = scipy.sparse.csc_matrix([[0., 1.], [1., 0.]], dtype=complex)
pauli_y_csc = scipy.sparse.csc_matrix([[0., -1.j], [1.j, 0.]], dtype=complex)
pauli_z_csc = scipy.sparse.csc_matrix([[1., 0.], [0., -1.]], dtype=complex)
q_raise_csc = (pauli_x_csc - 1.j * pauli_y_csc) / 2.
q_lower_csc = (pauli_x_csc + 1.j * pauli_y_csc) / 2.
pauli_matrix_map = {
    'I': identity_csc,
    'X': pauli_x_csc,
    'Y': pauli_y_csc,
    'Z': pauli_z_csc
}


def wrapped_kronecker(operator_1, operator_2):
    """Return the Kronecker product of two sparse.csc_matrix operators."""
    return scipy.sparse.kron(operator_1, operator_2, 'csc')


def kronecker_operators(*args):
    """Return the Kronecker product of multiple sparse.csc_matrix operators."""
    return reduce(wrapped_kronecker, *args)


def jordan_wigner_ladder_sparse(n_qubits, tensor_factor, ladder_type):
    r"""Make a matrix representation of a fermion ladder operator.

    Operators are mapped as follows:
    a_j^\dagger -> Z_0 .. Z_{j-1} (X_j - iY_j) / 2
    a_j -> Z_0 .. Z_{j-1} (X_j + iY_j) / 2

    Args:
        index: This is a nonzero integer. The integer indicates the tensor
            factor and the sign indicates raising or lowering.
        n_qubits(int): Number qubits in the system Hilbert space.

    Returns:
        The corresponding Scipy sparse matrix.
    """
    parities = tensor_factor * [pauli_z_csc]
    identities = [
        scipy.sparse.identity(2**(n_qubits - tensor_factor - 1),
                              dtype=complex,
                              format='csc')
    ]
    if ladder_type:
        operator = kronecker_operators(parities + [q_raise_csc] + identities)
    else:
        operator = kronecker_operators(parities + [q_lower_csc] + identities)
    return operator


def jordan_wigner_sparse(fermion_operator, n_qubits=None):
    r"""Initialize a Scipy sparse matrix from a FermionOperator.

    Operators are mapped as follows:
    a_j^\dagger -> Z_0 .. Z_{j-1} (X_j - iY_j) / 2
    a_j -> Z_0 .. Z_{j-1} (X_j + iY_j) / 2

    Args:
        fermion_operator(FermionOperator): instance of the FermionOperator
            class.
        n_qubits(int): Number of qubits.

    Returns:
        The corresponding Scipy sparse matrix.
    """
    if n_qubits is None:
        n_qubits = count_qubits(fermion_operator)

    # Create a list of raising and lowering operators for each orbital.
    jw_operators = []
    for tensor_factor in range(n_qubits):
        jw_operators += [(jordan_wigner_ladder_sparse(n_qubits, tensor_factor,
                                                      0),
                          jordan_wigner_ladder_sparse(n_qubits, tensor_factor,
                                                      1))]

    # Construct the Scipy sparse matrix.
    n_hilbert = 2**n_qubits
    values_list = [[]]
    row_list = [[]]
    column_list = [[]]
    for term in fermion_operator.terms:
        coefficient = fermion_operator.terms[term]
        sparse_matrix = coefficient * scipy.sparse.identity(
            2**n_qubits, dtype=complex, format='csc')
        for ladder_operator in term:
            sparse_matrix = sparse_matrix * jw_operators[ladder_operator[0]][
                ladder_operator[1]]

        if coefficient:
            # Extract triplets from sparse_term.
            sparse_matrix = sparse_matrix.tocoo(copy=False)
            values_list.append(sparse_matrix.data)
            (row, column) = sparse_matrix.nonzero()
            row_list.append(row)
            column_list.append(column)

    values_list = numpy.concatenate(values_list)
    row_list = numpy.concatenate(row_list)
    column_list = numpy.concatenate(column_list)
    sparse_operator = scipy.sparse.coo_matrix(
        (values_list, (row_list, column_list)),
        shape=(n_hilbert, n_hilbert)).tocsc(copy=False)
    sparse_operator.eliminate_zeros()
    return sparse_operator


def qubit_operator_sparse(qubit_operator, n_qubits=None):
    """Initialize a Scipy sparse matrix from a QubitOperator.

    Args:
        qubit_operator(QubitOperator): instance of the QubitOperator class.
        n_qubits (int): Number of qubits.

    Returns:
        The corresponding Scipy sparse matrix.
    """
    if n_qubits is None:
        n_qubits = count_qubits(qubit_operator)
    if n_qubits < count_qubits(qubit_operator):
        raise ValueError('Invalid number of qubits specified.')

    # Construct the Scipy sparse matrix.
    n_hilbert = 2**n_qubits
    values_list = [[]]
    row_list = [[]]
    column_list = [[]]

    # Loop through the terms.
    for qubit_term in qubit_operator.terms:
        tensor_factor = 0
        coefficient = qubit_operator.terms[qubit_term]
        sparse_operators = [coefficient]
        for pauli_operator in qubit_term:

            # Grow space for missing identity operators.
            if pauli_operator[0] > tensor_factor:
                identity_qubits = pauli_operator[0] - tensor_factor
                identity = scipy.sparse.identity(2**identity_qubits,
                                                 dtype=complex,
                                                 format='csc')
                sparse_operators += [identity]

            # Add actual operator to the list.
            sparse_operators += [pauli_matrix_map[pauli_operator[1]]]
            tensor_factor = pauli_operator[0] + 1

        # Grow space at end of string unless operator acted on final qubit.
        if tensor_factor < n_qubits or not qubit_term:
            identity_qubits = n_qubits - tensor_factor
            identity = scipy.sparse.identity(2**identity_qubits,
                                             dtype=complex,
                                             format='csc')
            sparse_operators += [identity]

        # Extract triplets from sparse_term.
        sparse_matrix = kronecker_operators(sparse_operators)
        values_list.append(sparse_matrix.tocoo(copy=False).data)
        (column, row) = sparse_matrix.nonzero()
        column_list.append(column)
        row_list.append(row)

    # Create sparse operator.
    values_list = numpy.concatenate(values_list)
    row_list = numpy.concatenate(row_list)
    column_list = numpy.concatenate(column_list)
    sparse_operator = scipy.sparse.coo_matrix(
        (values_list, (row_list, column_list)),
        shape=(n_hilbert, n_hilbert)).tocsc(copy=False)
    sparse_operator.eliminate_zeros()
    return sparse_operator


def get_linear_qubit_operator_diagonal(qubit_operator, n_qubits=None):
    """ Return a linear operator's diagonal elements.

    The main motivation is to use it for Davidson's algorithm, to find out the
    lowest n eigenvalues and associated eigenvectors.

    Qubit terms with X or Y operators will contribute nothing to the diagonal
    elements, while I or Z will contribute a factor of 1 or -1 together with
    the coefficient.

    Args:
        qubit_operator(QubitOperator): A qubit operator.

    Returns:
        linear_operator_diagonal(numpy.ndarray): The diagonal elements for
            LinearQubitOperator(qubit_operator).
    """
    if n_qubits is None:
        n_qubits = count_qubits(qubit_operator)
    if n_qubits < count_qubits(qubit_operator):
        raise ValueError('Invalid number of qubits specified.')

    n_hilbert = 2**n_qubits
    zeros_diagonal = numpy.zeros(n_hilbert)
    ones_diagonal = numpy.ones(n_hilbert)
    linear_operator_diagonal = zeros_diagonal
    # Loop through the terms.
    for qubit_term in qubit_operator.terms:
        is_zero = False
        tensor_factor = 0
        vecs = [ones_diagonal]
        for pauli_operator in qubit_term:
            op = pauli_operator[1]
            if op in ['X', 'Y']:
                is_zero = True
                break

            # Split vector by half and half for each bit.
            if pauli_operator[0] > tensor_factor:
                vecs = [
                    v for iter_v in vecs
                    for v in numpy.split(iter_v, 2**(pauli_operator[0] -
                                                     tensor_factor))
                ]

            vec_pairs = [numpy.split(v, 2) for v in vecs]
            vecs = [v for vp in vec_pairs for v in (vp[0], -vp[1])]
            tensor_factor = pauli_operator[0] + 1
        if not is_zero:
            linear_operator_diagonal += (qubit_operator.terms[qubit_term] *
                                         numpy.concatenate(vecs))
    return linear_operator_diagonal


def jw_configuration_state(occupied_orbitals, n_qubits):
    """Function to produce a basis state in the occupation number basis.

    Args:
        occupied_orbitals(list): A list of integers representing the indices
            of the occupied orbitals in the desired basis state
        n_qubits(int): The total number of qubits

    Returns:
        basis_vector(sparse): The basis state as a sparse matrix
    """
    one_index = sum(2**(n_qubits - 1 - i) for i in occupied_orbitals)
    basis_vector = numpy.zeros(2**n_qubits, dtype=float)
    basis_vector[one_index] = 1
    return basis_vector


def jw_hartree_fock_state(n_electrons, n_orbitals):
    """Function to produce Hartree-Fock state in JW representation."""
    hartree_fock_state = jw_configuration_state(range(n_electrons), n_orbitals)
    return hartree_fock_state


def jw_number_indices(n_electrons, n_qubits):
    """Return the indices for n_electrons in n_qubits under JW encoding

    Calculates the indices for all possible arrangements of n-electrons
        within n-qubit orbitals when a Jordan-Wigner encoding is used.
        Useful for restricting generic operators or vectors to a particular
        particle number space when desired

    Args:
        n_electrons(int): Number of particles to restrict the operator to
        n_qubits(int): Number of qubits defining the total state

    Returns:
        indices(list): List of indices in a 2^n length array that indicate
            the indices of constant particle number within n_qubits
            in a Jordan-Wigner encoding.
    """
    occupations = itertools.combinations(range(n_qubits), n_electrons)
    indices = [sum([2**n for n in occupation]) for occupation in occupations]
    return indices


def jw_sz_indices(sz_value,
                  n_qubits,
                  n_electrons=None,
                  up_index=up_index,
                  down_index=down_index):
    r"""Return the indices of basis vectors with fixed Sz under JW encoding.

    The returned indices label computational basis vectors which lie within
    the corresponding eigenspace of the Sz operator,

    .. math::
        \begin{align}
        S^{z} = \frac{1}{2}\sum_{i = 1}^{n}(n_{i, \alpha} - n_{i, \beta})
        \end{align}

    Args:
        sz_value(float): Desired Sz value. Should be an integer or
            half-integer.
        n_qubits(int): Number of qubits defining the total state
        n_electrons(int, optional): Number of particles to restrict the
            operator to, if such a restriction is desired
        up_index (Callable, optional): Function that maps a spatial index
            to the index of the corresponding up site
        down_index (Callable, optional): Function that maps a spatial index
            to the index of the corresponding down site

    Returns:
        indices(list): The list of indices
    """
    if n_qubits % 2 != 0:
        raise ValueError('Number of qubits must be even')

    if not (2. * sz_value).is_integer():
        raise ValueError('Sz value must be an integer or half-integer')

    n_sites = n_qubits // 2
    sz_integer = int(2. * sz_value)
    indices = []

    if n_electrons is not None:
        # Particle number is fixed, so the number of spin-up electrons
        # (as well as the number of spin-down electrons) is fixed
        if ((n_electrons + sz_integer) % 2 != 0 or
                n_electrons < abs(sz_integer)):
            raise ValueError('The specified particle number and sz value are '
                             'incompatible.')
        num_up = (n_electrons + sz_integer) // 2
        num_down = n_electrons - num_up
        up_occupations = itertools.combinations(range(n_sites), num_up)
        down_occupations = list(itertools.combinations(range(n_sites),
                                                       num_down))
        # Each arrangement of up spins can be paired with an arrangement
        # of down spins
        for up_occupation in up_occupations:
            up_occupation = [up_index(index) for index in up_occupation]
            for down_occupation in down_occupations:
                down_occupation = [
                    down_index(index) for index in down_occupation
                ]
                occupation = up_occupation + down_occupation
                indices.append(sum(2**(n_qubits - 1 - k) for k in occupation))
    else:
        # Particle number is not fixed
        if sz_integer < 0:
            # There are more down spins than up spins
            more_map = down_index
            less_map = up_index
        else:
            # There are at least as many up spins as down spins
            more_map = up_index
            less_map = down_index
        for n in range(abs(sz_integer), n_sites + 1):
            # Choose n of the 'more' spin and n - abs(sz_integer) of the
            # 'less' spin
            more_occupations = itertools.combinations(range(n_sites), n)
            less_occupations = list(
                itertools.combinations(range(n_sites), n - abs(sz_integer)))
            # Each arrangement of the 'more' spins can be paired with an
            # arrangement of the 'less' spin
            for more_occupation in more_occupations:
                more_occupation = [more_map(index) for index in more_occupation]
                for less_occupation in less_occupations:
                    less_occupation = [
                        less_map(index) for index in less_occupation
                    ]
                    occupation = more_occupation + less_occupation
                    indices.append(
                        sum(2**(n_qubits - 1 - k) for k in occupation))

    return indices


def jw_number_restrict_operator(operator, n_electrons, n_qubits=None):
    """Restrict a Jordan-Wigner encoded operator to a given particle number

    Args:
        sparse_operator(ndarray or sparse): Numpy operator acting on
            the space of n_qubits.
        n_electrons(int): Number of particles to restrict the operator to
        n_qubits(int): Number of qubits defining the total state

    Returns:
        new_operator(ndarray or sparse): Numpy operator restricted to
            acting on states with the same particle number.
    """
    if n_qubits is None:
        n_qubits = int(numpy.log2(operator.shape[0]))

    select_indices = jw_number_indices(n_electrons, n_qubits)
    return operator[numpy.ix_(select_indices, select_indices)]


def jw_sz_restrict_operator(operator,
                            sz_value,
                            n_electrons=None,
                            n_qubits=None,
                            up_index=up_index,
                            down_index=down_index):
    """Restrict a Jordan-Wigner encoded operator to a given Sz value

    Args:
        operator(ndarray or sparse): Numpy operator acting on
            the space of n_qubits.
        sz_value(float): Desired Sz value. Should be an integer or
            half-integer.
        n_electrons(int, optional): Number of particles to restrict the
            operator to, if such a restriction is desired.
        n_qubits(int, optional): Number of qubits defining the total state
        up_index (Callable, optional): Function that maps a spatial index
            to the index of the corresponding up site
        down_index (Callable, optional): Function that maps a spatial index
            to the index of the corresponding down site

    Returns:
        new_operator(ndarray or sparse): Numpy operator restricted to
            acting on states with the desired Sz value.
    """
    if n_qubits is None:
        n_qubits = int(numpy.log2(operator.shape[0]))

    select_indices = jw_sz_indices(sz_value,
                                   n_qubits,
                                   n_electrons=n_electrons,
                                   up_index=up_index,
                                   down_index=down_index)
    return operator[numpy.ix_(select_indices, select_indices)]


def jw_number_restrict_state(state, n_electrons, n_qubits=None):
    """Restrict a Jordan-Wigner encoded state to a given particle number

    Args:
        state(ndarray or sparse): Numpy vector in
            the space of n_qubits.
        n_electrons(int): Number of particles to restrict the state to
        n_qubits(int): Number of qubits defining the total state

    Returns:
        new_operator(ndarray or sparse): Numpy vector restricted to
            states with the same particle number. May not be normalized.
    """
    if n_qubits is None:
        n_qubits = int(numpy.log2(state.shape[0]))

    select_indices = jw_number_indices(n_electrons, n_qubits)
    return state[select_indices]


def jw_sz_restrict_state(state,
                         sz_value,
                         n_electrons=None,
                         n_qubits=None,
                         up_index=up_index,
                         down_index=down_index):
    """Restrict a Jordan-Wigner encoded state to a given Sz value

    Args:
        state(ndarray or sparse): Numpy vector in
            the space of n_qubits.
        sz_value(float): Desired Sz value. Should be an integer or
            half-integer.
        n_electrons(int, optional): Number of particles to restrict the
            operator to, if such a restriction is desired.
        n_qubits(int, optional): Number of qubits defining the total state
        up_index (Callable, optional): Function that maps a spatial index
            to the index of the corresponding up site
        down_index (Callable, optional): Function that maps a spatial index
            to the index of the corresponding down site

    Returns:
        new_operator(ndarray or sparse): Numpy vector restricted to
            states with the desired Sz value. May not be normalized.
    """
    if n_qubits is None:
        n_qubits = int(numpy.log2(state.shape[0]))

    select_indices = jw_sz_indices(sz_value,
                                   n_qubits,
                                   n_electrons=n_electrons,
                                   up_index=up_index,
                                   down_index=down_index)
    return state[select_indices]


def jw_get_ground_state_at_particle_number(sparse_operator, particle_number):
    """Compute ground energy and state at a specified particle number.

    Assumes the Jordan-Wigner transform. The input operator should be Hermitian
    and particle-number-conserving.

    Args:
        sparse_operator(sparse): A Jordan-Wigner encoded sparse matrix.
        particle_number(int): The particle number at which to compute the ground
            energy and states

    Returns:
        ground_energy(float): The lowest eigenvalue of sparse_operator within
            the eigenspace of the number operator corresponding to
            particle_number.
        ground_state(ndarray): The ground state at the particle number
    """

    n_qubits = int(numpy.log2(sparse_operator.shape[0]))

    # Get the operator restricted to the subspace of the desired particle number
    restricted_operator = jw_number_restrict_operator(sparse_operator,
                                                      particle_number, n_qubits)

    # Compute eigenvalues and eigenvectors
    if restricted_operator.shape[0] - 1 <= 1:
        # Restricted operator too small for sparse eigensolver
        dense_restricted_operator = restricted_operator.toarray()
        eigvals, eigvecs = numpy.linalg.eigh(dense_restricted_operator)
    else:
        eigvals, eigvecs = scipy.sparse.linalg.eigsh(restricted_operator,
                                                     k=1,
                                                     which='SA')

    # Expand the state
    state = eigvecs[:, 0]
    expanded_state = numpy.zeros(2**n_qubits, dtype=complex)
    expanded_state[jw_number_indices(particle_number, n_qubits)] = state

    return eigvals[0], expanded_state


def jw_sparse_givens_rotation(i, j, theta, phi, n_qubits):
    """Return the matrix (acting on a full wavefunction) that performs a
    Givens rotation of modes i and j in the Jordan-Wigner encoding."""
    if j != i + 1:
        raise ValueError('Only adjacent modes can be rotated.')
    if j > n_qubits - 1:
        raise ValueError('Too few qubits requested.')

    cosine = numpy.cos(theta)
    sine = numpy.sin(theta)
    phase = numpy.exp(1.j * phi)

    # Create the two-qubit rotation matrix
    rotation_matrix = scipy.sparse.csc_matrix(
        ([1., phase * cosine, -phase * sine, sine, cosine, phase],
         ((0, 1, 1, 2, 2, 3), (0, 1, 2, 1, 2, 3))),
        shape=(4, 4),
        dtype=numpy.complex128)

    # Initialize identity operators
    left_eye = scipy.sparse.eye(2**i, format='csc')
    right_eye = scipy.sparse.eye(2**(n_qubits - 1 - j), format='csc')

    # Construct the matrix and return
    givens_matrix = kronecker_operators([left_eye, rotation_matrix, right_eye])

    return givens_matrix


def jw_sparse_particle_hole_transformation_last_mode(n_qubits):
    """Return the matrix (acting on a full wavefunction) that performs a
    particle-hole transformation on the last mode in the Jordan-Wigner
    encoding.
    """
    left_eye = scipy.sparse.eye(2**(n_qubits - 1), format='csc')
    return kronecker_operators([left_eye, pauli_matrix_map['X']])


def get_density_matrix(states, probabilities):
    n_qubits = states[0].shape[0]
    density_matrix = scipy.sparse.csc_matrix((n_qubits, n_qubits),
                                             dtype=complex)
    for state, probability in zip(states, probabilities):
        state = scipy.sparse.csc_matrix(state.reshape((len(state), 1)))
        density_matrix = density_matrix + probability * state * state.getH()
    return density_matrix


def get_ground_state(sparse_operator, initial_guess=None):
    """Compute lowest eigenvalue and eigenstate.

    Args:
        sparse_operator (LinearOperator): Operator to find the ground state of.
        initial_guess (ndarray): Initial guess for ground state.  A good
            guess dramatically reduces the cost required to converge.

    Returns
    -------
        eigenvalue:
            The lowest eigenvalue, a float.
        eigenstate:
            The lowest eigenstate in scipy.sparse csc format.
    """
    values, vectors = scipy.sparse.linalg.eigsh(sparse_operator,
                                                k=1,
                                                v0=initial_guess,
                                                which='SA',
                                                maxiter=1e7)

    order = numpy.argsort(values)
    values = values[order]
    vectors = vectors[:, order]
    eigenvalue = values[0]
    eigenstate = vectors[:, 0]
    return eigenvalue, eigenstate.T


def eigenspectrum(operator, n_qubits=None):
    """Compute the eigenspectrum of an operator.

    WARNING: This function has cubic runtime in dimension of
        Hilbert space operator, which might be exponential.

    NOTE: This function does not currently support
        QuadOperator and BosonOperator.

    Args:
        operator: QubitOperator, InteractionOperator, FermionOperator,
            PolynomialTensor, or InteractionRDM.
        n_qubits (int): number of qubits/modes in operator. if None, will
            be counted.

    Returns:
        spectrum: dense numpy array of floats giving eigenspectrum.
    """
    if isinstance(operator, (QuadOperator, BosonOperator)):
        raise TypeError('Operator of invalid type.')
    sparse_operator = get_sparse_operator(operator, n_qubits)
    spectrum = sparse_eigenspectrum(sparse_operator)
    return spectrum


def sparse_eigenspectrum(sparse_operator):
    """Perform a dense diagonalization.

    Returns:
        eigenspectrum: The lowest eigenvalues in a numpy array.
    """
    dense_operator = sparse_operator.todense()
    if is_hermitian(sparse_operator):
        eigenspectrum = numpy.linalg.eigvalsh(dense_operator)
    else:
        eigenspectrum = numpy.linalg.eigvals(dense_operator)
    return numpy.sort(eigenspectrum)


def expectation(operator, state):
    """Compute the expectation value of an operator with a state.

    Args:
        operator(scipy.sparse.spmatrix or scipy.sparse.linalg.LinearOperator):
            The operator whose expectation value is desired.
        state(numpy.ndarray or scipy.sparse.spmatrix): A numpy array
            representing a pure state or a sparse matrix representing a density
            matrix. If `operator` is a LinearOperator, then this must be a
            numpy array.

    Returns:
        A complex number giving the expectation value.

    Raises:
        ValueError: Input state has invalid format.
    """

    if isinstance(state, scipy.sparse.spmatrix):
        # Handle density matrix.
        if isinstance(operator, scipy.sparse.linalg.LinearOperator):
            raise ValueError('Taking the expectation of a LinearOperator with '
                             'a density matrix is not supported.')
        product = state * operator
        expectation = numpy.sum(product.diagonal())

    elif isinstance(state, numpy.ndarray):
        # Handle state vector.
        if len(state.shape) == 1:
            # Row vector
            expectation = numpy.dot(numpy.conjugate(state), operator * state)
        else:
            # Column vector
            expectation = numpy.dot(numpy.conjugate(state.T),
                                    operator * state)[0, 0]

    else:
        # Handle exception.
        raise ValueError(
            'Input state must be a numpy array or a sparse matrix.')

    # Return.
    return expectation


def variance(operator, state):
    """Compute variance of operator with a state.

    Args:
        operator(scipy.sparse.spmatrix or scipy.sparse.linalg.LinearOperator):
            The operator whose expectation value is desired.
        state(numpy.ndarray or scipy.sparse.spmatrix): A numpy array
            representing a pure state or a sparse matrix representing a density
            matrix.

    Returns:
        A complex number giving the variance.

    Raises:
        ValueError: Input state has invalid format.
    """
    return (expectation(operator**2, state) - expectation(operator, state)**2)


def expectation_computational_basis_state(operator, computational_basis_state):
    """Compute expectation value of operator with a  state.

    Args:
        operator: Qubit or FermionOperator to evaluate expectation value of.
                  If operator is a FermionOperator, it must be normal-ordered.
        computational_basis_state (scipy.sparse vector / list): normalized
            computational basis state (if scipy.sparse vector), or list of
            occupied orbitals.

    Returns:
        A real float giving expectation value.

    Raises:
        TypeError: Incorrect operator or state type.
    """
    if isinstance(operator, QubitOperator):
        raise NotImplementedError('Not yet implemented for QubitOperators.')

    if not isinstance(operator, FermionOperator):
        raise TypeError('operator must be a FermionOperator.')

    occupied_orbitals = computational_basis_state

    if not isinstance(occupied_orbitals, list):
        computational_basis_state_index = (occupied_orbitals.nonzero()[0][0])

        occupied_orbitals = [
            digit == '1' for digit in bin(computational_basis_state_index)[2:]
        ][::-1]

    expectation_value = operator.terms.get((), 0.0)

    for i in range(len(occupied_orbitals)):
        if occupied_orbitals[i]:
            expectation_value += operator.terms.get(((i, 1), (i, 0)), 0.0)

            for j in range(i + 1, len(occupied_orbitals)):
                expectation_value -= operator.terms.get(
                    ((j, 1), (i, 1), (j, 0), (i, 0)), 0.0)

    return expectation_value


def expectation_db_operator_with_pw_basis_state(operator,
                                                plane_wave_occ_orbitals,
                                                n_spatial_orbitals, grid,
                                                spinless):
    """Compute expectation value of a dual basis operator with a plane
    wave computational basis state.

    Args:
        operator: Dual-basis representation of FermionOperator to evaluate
                  expectation value of. Can have at most 3-body terms.
        plane_wave_occ_orbitals (list): list of occupied plane-wave orbitals.
        n_spatial_orbitals (int): Number of spatial orbitals.
        grid (openfermion.utils.Grid): The grid used for discretization.
        spinless (bool): Whether the system is spinless.

    Returns:
        A real float giving the expectation value.
    """
    expectation_value = operator.terms.get((), 0.0)

    for single_action, coefficient in operator.terms.items():
        if len(single_action) == 2:
            expectation_value += coefficient * (
                expectation_one_body_db_operator_computational_basis_state(
                    single_action, plane_wave_occ_orbitals, grid, spinless) /
                n_spatial_orbitals)

        elif len(single_action) == 4:
            expectation_value += coefficient * (
                expectation_two_body_db_operator_computational_basis_state(
                    single_action, plane_wave_occ_orbitals, grid, spinless) /
                n_spatial_orbitals**2)

        elif len(single_action) == 6:
            expectation_value += coefficient * (
                expectation_three_body_db_operator_computational_basis_state(
                    single_action, plane_wave_occ_orbitals, grid, spinless) /
                n_spatial_orbitals**3)

    return expectation_value


def expectation_one_body_db_operator_computational_basis_state(
        dual_basis_action, plane_wave_occ_orbitals, grid, spinless):
    """Compute expectation value of a 1-body dual-basis operator with a
    plane wave computational basis state.

    Args:
        dual_basis_action: Dual-basis action of FermionOperator to
                           evaluate expectation value of.
        plane_wave_occ_orbitals (list): list of occupied plane-wave orbitals.
        grid (openfermion.utils.Grid): The grid used for discretization.
        spinless (bool): Whether the system is spinless.

    Returns:
        A real float giving the expectation value.
    """
    expectation_value = 0.0

    r_p = grid.position_vector(
        grid.grid_indices(dual_basis_action[0][0], spinless))
    r_q = grid.position_vector(
        grid.grid_indices(dual_basis_action[1][0], spinless))

    for orbital in plane_wave_occ_orbitals:
        # If there's spin, p and q have to have the same parity (spin),
        # and the new orbital has to have the same spin as these.
        k_orbital = grid.momentum_vector(grid.grid_indices(orbital, spinless))
        # The Fourier transform is spin-conserving. This means that p, q,
        # and the new orbital all have to have the same spin (parity).
        if spinless or (dual_basis_action[0][0] % 2 ==
                        dual_basis_action[1][0] % 2 == orbital % 2):
            expectation_value += numpy.exp(-1j * k_orbital.dot(r_p - r_q))

    return expectation_value


def expectation_two_body_db_operator_computational_basis_state(
        dual_basis_action, plane_wave_occ_orbitals, grid, spinless):
    """Compute expectation value of a 2-body dual-basis operator with a
    plane wave computational basis state.

    Args:
        dual_basis_action: Dual-basis action of FermionOperator to
                           evaluate expectation value of.
        plane_wave_occ_orbitals (list): list of occupied plane-wave orbitals.
        grid (openfermion.utils.Grid): The grid used for discretization.
        spinless (bool): Whether the system is spinless.

    Returns:
        A float giving the expectation value.
    """
    expectation_value = 0.0

    r = {}
    for i in range(4):
        r[i] = grid.position_vector(
            grid.grid_indices(dual_basis_action[i][0], spinless))

    rr = {}
    k_map = {}
    for i in range(2):
        rr[i] = {}
        k_map[i] = {}
        for j in range(2, 4):
            rr[i][j] = r[i] - r[j]
            k_map[i][j] = {}

    # Pre-computations.
    for o in plane_wave_occ_orbitals:
        k = grid.momentum_vector(grid.grid_indices(o, spinless))
        for i in range(2):
            for j in range(2, 4):
                k_map[i][j][o] = k.dot(rr[i][j])

    for orbital1 in plane_wave_occ_orbitals:
        k1ac = k_map[0][2][orbital1]
        k1ad = k_map[0][3][orbital1]

        for orbital2 in plane_wave_occ_orbitals:
            if orbital1 != orbital2:
                k2bc = k_map[1][2][orbital2]
                k2bd = k_map[1][3][orbital2]

                # The Fourier transform is spin-conserving. This means that
                # the parity of the orbitals involved in the transition must
                # be the same.
                if spinless or (
                    (dual_basis_action[0][0] % 2 == dual_basis_action[3][0] % 2
                     == orbital1 % 2) and
                    (dual_basis_action[1][0] % 2 == dual_basis_action[2][0] % 2
                     == orbital2 % 2)):
                    value = numpy.exp(-1j * (k1ad + k2bc))

                    # Add because it came from two anti-commutations.
                    expectation_value += value

                # The Fourier transform is spin-conserving. This means that
                # the parity of the orbitals involved in the transition must
                # be the same.
                if spinless or (
                    (dual_basis_action[0][0] % 2 == dual_basis_action[2][0] % 2
                     == orbital1 % 2) and
                    (dual_basis_action[1][0] % 2 == dual_basis_action[3][0] % 2
                     == orbital2 % 2)):
                    value = numpy.exp(-1j * (k1ac + k2bd))

                    # Subtract because it came from a single anti-commutation.
                    expectation_value -= value

    return expectation_value


def expectation_three_body_db_operator_computational_basis_state(
        dual_basis_action, plane_wave_occ_orbitals, grid, spinless):
    """Compute expectation value of a 3-body dual-basis operator with a
    plane wave computational basis state.

    Args:
        dual_basis_action: Dual-basis action of FermionOperator to
                           evaluate expectation value of.
        plane_wave_occ_orbitals (list): list of occupied plane-wave orbitals.
        grid (openfermion.utils.Grid): The grid used for discretization.
        spinless (bool): Whether the system is spinless.

    Returns:
        A float giving the expectation value.
    """
    expectation_value = 0.0

    r = {}
    for i in range(6):
        r[i] = grid.position_vector(
            grid.grid_indices(dual_basis_action[i][0], spinless))

    rr = {}
    k_map = {}
    for i in range(3):
        rr[i] = {}
        k_map[i] = {}
        for j in range(3, 6):
            rr[i][j] = r[i] - r[j]
            k_map[i][j] = {}

    # Pre-computations.
    for o in plane_wave_occ_orbitals:
        k = grid.momentum_vector(grid.grid_indices(o, spinless))
        for i in range(3):
            for j in range(3, 6):
                k_map[i][j][o] = k.dot(rr[i][j])

    for orbital1 in plane_wave_occ_orbitals:
        k1ad = k_map[0][3][orbital1]
        k1ae = k_map[0][4][orbital1]
        k1af = k_map[0][5][orbital1]

        for orbital2 in plane_wave_occ_orbitals:
            if orbital1 != orbital2:
                k2bd = k_map[1][3][orbital2]
                k2be = k_map[1][4][orbital2]
                k2bf = k_map[1][5][orbital2]

                for orbital3 in plane_wave_occ_orbitals:
                    if orbital1 != orbital3 and orbital2 != orbital3:
                        k3cd = k_map[2][3][orbital3]
                        k3ce = k_map[2][4][orbital3]
                        k3cf = k_map[2][5][orbital3]

                        # Handle \delta_{ad} \delta_{bf} \delta_{ce} after FT.
                        # The Fourier transform is spin-conserving.
                        if spinless or (
                            (dual_basis_action[0][0] % 2 ==
                             dual_basis_action[3][0] % 2 == orbital1 % 2) and
                            (dual_basis_action[1][0] % 2 ==
                             dual_basis_action[5][0] % 2 == orbital2 % 2) and
                            (dual_basis_action[2][0] % 2 ==
                             dual_basis_action[4][0] % 2 == orbital3 % 2)):
                            expectation_value += numpy.exp(-1j *
                                                           (k1ad + k2bf + k3ce))

                        # Handle -\delta_{ad} \delta_{be} \delta_{cf} after FT.
                        # The Fourier transform is spin-conserving.
                        if spinless or (
                            (dual_basis_action[0][0] % 2 ==
                             dual_basis_action[3][0] % 2 == orbital1 % 2) and
                            (dual_basis_action[1][0] % 2 ==
                             dual_basis_action[4][0] % 2 == orbital2 % 2) and
                            (dual_basis_action[2][0] % 2 ==
                             dual_basis_action[5][0] % 2 == orbital3 % 2)):
                            expectation_value -= numpy.exp(-1j *
                                                           (k1ad + k2be + k3cf))

                        # Handle -\delta_{ae} \delta_{bf} \delta_{cd} after FT.
                        # The Fourier transform is spin-conserving.
                        if spinless or (
                            (dual_basis_action[0][0] % 2 ==
                             dual_basis_action[4][0] % 2 == orbital1 % 2) and
                            (dual_basis_action[1][0] % 2 ==
                             dual_basis_action[5][0] % 2 == orbital2 % 2) and
                            (dual_basis_action[2][0] % 2 ==
                             dual_basis_action[3][0] % 2 == orbital3 % 2)):
                            expectation_value -= numpy.exp(-1j *
                                                           (k1ae + k2bf + k3cd))

                        # Handle \delta_{ae} \delta_{bd} \delta_{cf} after FT.
                        # The Fourier transform is spin-conserving.
                        if spinless or (
                            (dual_basis_action[0][0] % 2 ==
                             dual_basis_action[4][0] % 2 == orbital1 % 2) and
                            (dual_basis_action[1][0] % 2 ==
                             dual_basis_action[3][0] % 2 == orbital2 % 2) and
                            (dual_basis_action[2][0] % 2 ==
                             dual_basis_action[5][0] % 2 == orbital3 % 2)):
                            expectation_value += numpy.exp(-1j *
                                                           (k1ae + k2bd + k3cf))

                        # Handle \delta_{af} \delta_{be} \delta_{cd} after FT.
                        # The Fourier transform is spin-conserving.
                        if spinless or (
                            (dual_basis_action[0][0] % 2 ==
                             dual_basis_action[5][0] % 2 == orbital1 % 2) and
                            (dual_basis_action[1][0] % 2 ==
                             dual_basis_action[4][0] % 2 == orbital2 % 2) and
                            (dual_basis_action[2][0] % 2 ==
                             dual_basis_action[3][0] % 2 == orbital3 % 2)):
                            expectation_value += numpy.exp(-1j *
                                                           (k1af + k2be + k3cd))

                        # Handle -\delta_{af} \delta_{bd} \delta_{ce} after FT.
                        # The Fourier transform is spin-conserving.
                        if spinless or (
                            (dual_basis_action[0][0] % 2 ==
                             dual_basis_action[5][0] % 2 == orbital1 % 2) and
                            (dual_basis_action[1][0] % 2 ==
                             dual_basis_action[3][0] % 2 == orbital2 % 2) and
                            (dual_basis_action[2][0] % 2 ==
                             dual_basis_action[4][0] % 2 == orbital3 % 2)):
                            expectation_value -= numpy.exp(-1j *
                                                           (k1af + k2bd + k3ce))

    return expectation_value


def get_gap(sparse_operator, initial_guess=None):
    """Compute gap between lowest eigenvalue and first excited state.

    Args:
        sparse_operator (LinearOperator): Operator to find the ground state of.
        initial_guess (ndarray): Initial guess for eigenspace.  A good
            guess dramatically reduces the cost required to converge.
    Returns: A real float giving eigenvalue gap.
    """
    if not is_hermitian(sparse_operator):
        raise ValueError('sparse_operator must be Hermitian.')

    values, _ = scipy.sparse.linalg.eigsh(sparse_operator,
                                          k=2,
                                          v0=initial_guess,
                                          which='SA',
                                          maxiter=1e7)

    gap = abs(values[1] - values[0])
    return gap


def inner_product(state_1, state_2):
    """Compute inner product of two states."""
    return numpy.dot(state_1.conjugate(), state_2)


def boson_ladder_sparse(n_modes, mode, ladder_type, trunc):
    r"""Make a matrix representation of a singular bosonic ladder operator
    in the Fock space.

    Since the bosonic operator lies in an infinite Fock space,
    a truncation value needs to be provide so that a sparse matrix
    of finite size can be returned.

    Args:
        n_modes (int): Number of modes in the system Hilbert space.
        mode (int): The mode the ladder operator targets.
        ladder_type (int): This is a nonzero integer. 0 indicates a lowering
            operator, 1 a raising operator.
        trunc (int): The size at which the Fock space should be truncated
            when returning the matrix representing the ladder operator.

    Returns:
        The corresponding trunc x trunc Scipy sparse matrix.
    """
    if trunc < 1 or not isinstance(trunc, int):
        raise ValueError("Fock space truncation must be a positive integer.")

    if ladder_type:
        lop = scipy.sparse.spdiags(numpy.sqrt(range(1, trunc)),
                                   -1,
                                   trunc,
                                   trunc,
                                   format='csc')
    else:
        lop = scipy.sparse.spdiags(numpy.sqrt(range(trunc)),
                                   1,
                                   trunc,
                                   trunc,
                                   format='csc')

    Id = [scipy.sparse.identity(trunc, format='csc', dtype=complex)]
    operator_list = Id * mode + [lop] + Id * (n_modes - mode - 1)
    operator = kronecker_operators(operator_list)

    return operator


def single_quad_op_sparse(n_modes, mode, quadrature, hbar, trunc):
    r"""Make a matrix representation of a singular quadrature
    operator in the Fock space.

    Since the bosonic operators lie in an infinite Fock space,
    a truncation value needs to be provide so that a sparse matrix
    of finite size can be returned.

    Args:
        n_modes (int): Number of modes in the system Hilbert space.
        mode (int): The mode the ladder operator targets.
        quadrature (str): 'q' for the canonical position operator,
            'p' for the canonical moment]um operator.
        hbar (float): the value of hbar to use in the definition of the
            canonical commutation relation [q_i, p_j] = \delta_{ij} i hbar.
        trunc (int): The size at which the Fock space should be truncated
            when returning the matrix representing the ladder operator.

    Returns:
        The corresponding trunc x trunc Scipy sparse matrix.
    """
    if trunc < 1 or not isinstance(trunc, int):
        raise ValueError("Fock space truncation must be a positive integer.")

    b = boson_ladder_sparse(1, 0, 0, trunc)

    if quadrature == 'q':
        op = numpy.sqrt(hbar / 2) * (b + b.conj().T)
    elif quadrature == 'p':
        op = -1j * numpy.sqrt(hbar / 2) * (b - b.conj().T)

    Id = [scipy.sparse.identity(trunc, dtype=complex, format='csc')]
    operator_list = Id * mode + [op] + Id * (n_modes - mode - 1)
    operator = kronecker_operators(operator_list)

    return operator


def boson_operator_sparse(operator, trunc, hbar=1.):
    r"""Initialize a Scipy sparse matrix in the Fock space
    from a bosonic operator.

    Since the bosonic operators lie in an infinite Fock space,
    a truncation value needs to be provide so that a sparse matrix
    of finite size can be returned.

    Args:
        operator: One of either BosonOperator or QuadOperator.
        trunc (int): The size at which the Fock space should be truncated
            when returning the matrix representing the ladder operator.
        hbar (float): the value of hbar to use in the definition of the
            canonical commutation relation [q_i, p_j] = \delta_{ij} i hbar.
            This only applies if calcualating the sparse representation of
            a quadrature operator.

    Returns:
        The corresponding Scipy sparse matrix of size [trunc, trunc].
    """
    if isinstance(operator, QuadOperator):
        from openfermion.transforms.opconversions import get_boson_operator
        boson_operator = get_boson_operator(operator, hbar)
    elif isinstance(operator, BosonOperator):
        boson_operator = operator
    else:
        raise ValueError("Only BosonOperator and QuadOperator are supported.")

    if trunc < 1 or not isinstance(trunc, int):
        raise ValueError("Fock space truncation must be a positive integer.")

    # count the number of modes
    n_modes = 0
    for term in boson_operator.terms:
        for ladder_operator in term:
            if ladder_operator[0] + 1 > n_modes:
                n_modes = ladder_operator[0] + 1

    # Construct the Scipy sparse matrix.
    n_hilbert = trunc**n_modes
    values_list = [[]]
    row_list = [[]]
    column_list = [[]]

    # Loop through the terms.
    for term in boson_operator.terms:
        coefficient = boson_operator.terms[term]
        term_operator = coefficient * scipy.sparse.identity(
            n_hilbert, dtype=complex, format='csc')

        for ladder_op in term:
            # Add actual operator to the list.
            b = boson_ladder_sparse(n_modes, ladder_op[0], ladder_op[1], trunc)
            term_operator = term_operator.dot(b)

        # Extract triplets from sparse_term.
        values_list.append(term_operator.tocoo(copy=False).data)
        (row, column) = term_operator.nonzero()
        column_list.append(column)
        row_list.append(row)

    # Create sparse operator.
    values_list = numpy.concatenate(values_list)
    row_list = numpy.concatenate(row_list)
    column_list = numpy.concatenate(column_list)
    sparse_operator = scipy.sparse.coo_matrix(
        (values_list, (row_list, column_list)),
        shape=(n_hilbert, n_hilbert)).tocsc(copy=False)
    sparse_operator.eliminate_zeros()
    return sparse_operator


def get_sparse_operator(operator, n_qubits=None, trunc=None, hbar=1.):
    r"""Map an operator to a sparse matrix.

    If the input is not a QubitOperator, the Jordan-Wigner Transform is used.

    Args:
        operator: Currently supported operators include:
            FermionOperator, QubitOperator, DiagonalCoulombHamiltonian,
            PolynomialTensor, BosonOperator, QuadOperator.
        n_qubits(int): Number qubits in the system Hilbert space.
            Applicable only to fermionic systems.
        trunc (int): The size at which the Fock space should be truncated.
            Applicable only to bosonic systems.
        hbar (float): the value of hbar to use in the definition of the
            canonical commutation relation [q_i, p_j] = \delta_{ij} i hbar.
            Applicable only to the QuadOperator.
    """
    from openfermion.transforms.opconversions import get_fermion_operator
    if isinstance(operator, (DiagonalCoulombHamiltonian, PolynomialTensor)):
        return jordan_wigner_sparse(get_fermion_operator(operator))
    elif isinstance(operator, FermionOperator):
        return jordan_wigner_sparse(operator, n_qubits)
    elif isinstance(operator, QubitOperator):
        return qubit_operator_sparse(operator, n_qubits)
    elif isinstance(operator, (BosonOperator, QuadOperator)):
        return boson_operator_sparse(operator, trunc, hbar)
    else:
        raise TypeError('Failed to convert a {} to a sparse matrix.'.format(
            type(operator).__name__))


def get_number_preserving_sparse_operator(fermion_op,
                                          num_qubits,
                                          num_electrons,
                                          spin_preserving=False,
                                          reference_determinant=None,
                                          excitation_level=None):
    """Initialize a Scipy sparse matrix in a specific symmetry sector.

    This method initializes a Scipy sparse matrix from a FermionOperator,
    explicitly working in a particular particle number sector. Optionally, it
    can also restrict the space to contain only states with a particular Sz.

    Finally, the Hilbert space can also be restricted to only those states
    which are reachable by excitations up to a fixed rank from an initial
    reference determinant.

    Args:
        fermion_op(FermionOperator): An instance of the FermionOperator class.
            It should not contain terms which do not preserve particle number.
            If spin_preserving is set to True it should also not contain terms
            which do not preserve the Sz (it is assumed that the ordering of
            the indices goes alpha, beta, alpha, beta, ...).
        num_qubits(int): The total number of qubits / spin-orbitals in the
            system.
        num_electrons(int): The number of particles in the desired Hilbert
            space.
        spin_preserving(bool): Whether or not the constructed operator should
            be defined in a space which has support only on states with the
            same Sz value as the reference_determinant.
        reference_determinant(list(bool)): A list, whose length is equal to
            num_qubits, which specifies which orbitals should be occupied in
            the reference state. If spin_preserving is set to True then the Sz
            value of this reference state determines the Sz value of the
            symmetry sector in which the generated operator acts. If a value
            for excitation_level is provided then the excitations are generated
            with respect to the reference state. In any case, the ordering of
            the states in the matrix representation of the operator depends on
            reference_determinant and the state corresponding to
            reference_determinant is the vector [1.0, 0.0, 0.0 ... 0.0]. Can be
            set to None in order to take the first num_electrons orbitals to be
            the occupied orbitals.
        excitation_level(int): The number of excitations from the reference
            state which should be included in the generated operator's matrix
            representation. Can be set to None to include all levels of
            excitation.

    Returns:
        sparse_op(scipy.sparse.csc_matrix): A sparse matrix representation of
            fermion_op in the basis set by the arguments.
    """

    # We use the Hartree-Fock determinant as a reference if none is provided.
    if reference_determinant is None:
        reference_determinant = numpy.array(
            [i < num_electrons for i in range(num_qubits)])
    else:
        reference_determinant = numpy.asarray(reference_determinant)

    if excitation_level is None:
        excitation_level = num_electrons

    state_array = numpy.asarray(
        list(
            _iterate_basis_(reference_determinant, excitation_level,
                            spin_preserving)))
    # Create a 1d array with each determinant encoded
    # as an integer for sorting purposes.
    int_state_array = state_array.dot(
        1 << numpy.arange(state_array.shape[1])[::-1])
    sorting_indices = numpy.argsort(int_state_array)

    space_size = state_array.shape[0]

    fermion_op = normal_ordered(fermion_op)

    sparse_op = scipy.sparse.csc_matrix((space_size, space_size), dtype=float)

    for term, coefficient in fermion_op.terms.items():
        if len(term) == 0:
            constant = coefficient * scipy.sparse.identity(
                space_size, dtype=float, format='csc')

            sparse_op += constant

        else:
            term_op = _build_term_op_(term, state_array, int_state_array,
                                      sorting_indices)

            sparse_op += coefficient * term_op

    return sparse_op


def _iterate_basis_(reference_determinant, excitation_level, spin_preserving):
    """A helper method which iterates over the specified basis states.

    Note that this method always yields the states in order of their excitation
    rank from the reference_determinant.

    Args:
        reference_determinant(list(bool)): A list of bools which indicates
            which orbitals are occupied and which are unoccupied in the
            reference state.
        excitation_level(int): The maximum excitation rank to iterate over.
        spin_preserving(bool): A bool which, if set to True, constrains the
            method to iterate over only those states which have the same Sz as
            reference_determinant.

    Yields:
        Lists of bools which indicate which orbitals are occupied and which are
            unoccupied in the current determinant.
    """
    if not spin_preserving:
        for order in range(excitation_level + 1):
            for determinant in _iterate_basis_order_(reference_determinant,
                                                     order):
                yield determinant

    else:
        alpha_excitation_level = min(
            (numpy.sum(reference_determinant[::2]), excitation_level))
        beta_excitation_level = min(
            (numpy.sum(reference_determinant[1::2]), excitation_level))

        for order in range(excitation_level + 1):
            for alpha_order in range(alpha_excitation_level + 1):
                beta_order = order - alpha_order
                if (beta_order < 0 or beta_order > beta_excitation_level):
                    continue

                for determinant in _iterate_basis_spin_order_(
                        reference_determinant, alpha_order, beta_order):
                    yield determinant


def _iterate_basis_order_(reference_determinant, order):
    """A helper for iterating over determinants of a fixed excitation rank.

    Args:
        reference_determinant(list(bool)): The reference state with respect to
            which we are iterating over excited determinants.
        order(int): The number of excitations from the modes which are occupied
            in the reference_determinant.

    Yields:
        Lists of bools which indicate which orbitals are occupied and which are
            unoccupied in the current determinant.
        """
    occupied_indices = numpy.where(reference_determinant)[0]
    unoccupied_indices = numpy.where(numpy.invert(reference_determinant))[0]

    for occ_ind, unocc_ind in itertools.product(
            itertools.combinations(occupied_indices, order),
            itertools.combinations(unoccupied_indices, order)):
        basis_state = reference_determinant.copy()

        occ_ind = list(occ_ind)
        unocc_ind = list(unocc_ind)

        basis_state[occ_ind] = False
        basis_state[unocc_ind] = True

        yield basis_state


def _iterate_basis_spin_order_(reference_determinant, alpha_order, beta_order):
    """Iterates over states with a fixed excitation rank for each spin sector.

    This helper method assumes that the two spin sectors are interleaved:
    [1_alpha, 1_beta, 2_alpha, 2_beta, ...].

    Args:
        reference_determinant(list(bool)): The reference state with respect to
            which we are iterating over excited determinants.
        alpha_order(int): The number of excitations from the alpha spin sector
            of the reference_determinant.
        beta_order(int): The number of excitations from the beta spin sector of
            the reference_determinant.

    Yields:
        Lists of bools which indicate which orbitals are occupied and which are
            unoccupied in the current determinant.
        """
    occupied_alpha_indices = numpy.where(reference_determinant[::2])[0] * 2
    unoccupied_alpha_indices = numpy.where(
        numpy.invert(reference_determinant[::2]))[0] * 2
    occupied_beta_indices = numpy.where(reference_determinant[1::2])[0] * 2 + 1
    unoccupied_beta_indices = numpy.where(
        numpy.invert(reference_determinant[1::2]))[0] * 2 + 1

    for (alpha_occ_ind, alpha_unocc_ind, beta_occ_ind,
         beta_unocc_ind) in itertools.product(
             itertools.combinations(occupied_alpha_indices, alpha_order),
             itertools.combinations(unoccupied_alpha_indices, alpha_order),
             itertools.combinations(occupied_beta_indices, beta_order),
             itertools.combinations(unoccupied_beta_indices, beta_order)):
        basis_state = reference_determinant.copy()

        alpha_occ_ind = list(alpha_occ_ind)
        alpha_unocc_ind = list(alpha_unocc_ind)
        beta_occ_ind = list(beta_occ_ind)
        beta_unocc_ind = list(beta_unocc_ind)

        basis_state[alpha_occ_ind] = False
        basis_state[alpha_unocc_ind] = True
        basis_state[beta_occ_ind] = False
        basis_state[beta_unocc_ind] = True

        yield basis_state


def _build_term_op_(term, state_array, int_state_array, sorting_indices):
    """Builds a scipy sparse representation of a term from a FermionOperator.

    Args:
        term(tuple of tuple(int, int)s): The argument is a tuple of tuples
            representing a product of normal ordered fermionic creation and
            annihilation operators, each of which is of the form (int, int)
            where the first int indicates which site the operator acts on and
            the second int indicates whether the operator is a creation
            operator (1) or an annihilation operator (0). See the
            implementation of FermionOperator for more details.
        state_array(ndarray(bool)): A Numpy array which encodes each of the
            determinants in the space we are working in a bools which indicate
            the occupation of each mode. See the implementation of
            get_number_preserving_sparse_operator for more details.
        int_state_array(ndarray(int)): A one dimensional Numpy array which
            encodes the intFeger representation of the binary number
            corresponding to each determinant in state_array.
        sorting_indices(ndarray.view): A Numpy view which sorts
            int_state_array. This, together with int_state_array, allows for a
            quick lookup of the position of a particular determinant in
            state_array by converting it to its integer representation and
            searching through the sorted int_state_array.

    Raises:
        ValueError: If term does not represent a particle number conserving
            operator.

    Returns:
        A scipy.sparse.csc_matrix which corresponds to the operator specified
            by term expressed in the basis corresponding to the other arguments
            of the method."""

    space_size = state_array.shape[0]

    needs_to_be_occupied = []
    needs_to_be_unoccupied = []

    # We keep track of the number of creation and annihilation operators and
    # ensure that there are an equal number of them in order to help detect
    # invalid inputs.
    delta = 0
    for index, op_type in reversed(term):
        if op_type == 0:
            needs_to_be_occupied.append(index)
            delta -= 1
        else:
            if index not in needs_to_be_occupied:
                needs_to_be_unoccupied.append(index)
            delta += 1

    if delta != 0:
        raise ValueError(
            "The supplied operator doesn't preserve particle number")

    # We search for every state which has the necessary orbitals occupied and
    # unoccupied in order to not be immediately zeroed out based on the
    # creation and annihilation operators specified in term.
    maybe_valid_states = numpy.where(
        numpy.logical_and(
            numpy.all(state_array[:, needs_to_be_occupied], axis=1),
            numpy.logical_not(
                numpy.any(state_array[:, needs_to_be_unoccupied], axis=1))))[0]

    data = []
    row_ind = []
    col_ind = []
    shape = (space_size, space_size)

    # For each state that is not immediately zeroed out by the action of our
    # operator we check to see if the determinant which this state gets mapped
    # to is in the space we are considering.
    # Note that a failure to find any state does not necessarily indicate that
    # term specifies an invalid operator. For example, if we are restricting
    # ourselves to double excitations from a fixed reference state then the
    # action of term on some of our basis states may lead to determinants with
    # more than two excitations from the reference. These more than double
    # excited determinants are not included in the matrix representation (and
    # hence, will not be present in state_array).
    for _, state in enumerate(maybe_valid_states):
        determinant = state_array[state, :]
        target_determinant = determinant.copy()

        parity = 1
        for i, _ in reversed(term):
            area_to_check = target_determinant[0:i]
            parity *= (-1)**numpy.sum(area_to_check)

            target_determinant[i] = not target_determinant[i]

        int_encoding = target_determinant.dot(
            1 << numpy.arange(target_determinant.size)[::-1])

        target_state_index_sorted = numpy.searchsorted(int_state_array,
                                                       int_encoding,
                                                       sorter=sorting_indices)

        target_state = sorting_indices[target_state_index_sorted]

        if int_state_array[target_state] == int_encoding:
            # Then target state is in the space considered:
            data.append(parity)
            row_ind.append(target_state)
            col_ind.append(state)

    data = numpy.asarray(data)
    row_ind = numpy.asarray(row_ind)
    col_ind = numpy.asarray(col_ind)

    term_op = scipy.sparse.csc_matrix((data, (row_ind, col_ind)), shape=shape)

    return term_op



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Weyl ordering on bosonic operators."""

from scipy.special import binom

from openfermion.ops.operators import BosonOperator, QuadOperator


def mccoy(mode, op_a, op_b, m, n):
    """ Implement the McCoy formula on two operators of the
    form op_a^m op_b^n.

    Args:
        mode (int): the mode number the two operators act on.
        op_a: the label of operator a. This can be any hashable type.
        op_b: the label of operator b. This can be any hashable type.
        m (int): the power of operator a.
        n (int): the power of operator b.
    """
    new_op = dict()
    for r in range(0, n + 1):
        coeff = binom(n, r) / (2**n)
        new_term = tuple([(mode, op_b)] * r + [(mode, op_a)] * m +
                         [(mode, op_b)] * (n - r))
        if new_term not in new_op:
            new_op[tuple(new_term)] = coeff
        else:
            new_op[tuple(new_term)] += coeff
    return new_op


def weyl_polynomial_quantization(polynomial):
    r""" Apply the Weyl quantization to a phase space polynomial.

    The Weyl quantization is performed by applying McCoy's formula
    directly to a polynomial term of the form q^m p^n:

    q^m p^n ->
        (1/ 2^n) sum_{r=0}^{n} Binomial(n, r) \hat{q}^r \hat{p}^m q^{n-r}

    where q and p are phase space variables, and \hat{q} and \hat{p}
    are quadrature operators.

    The input is provided in the form of a string, for example

    .. code-block:: python

        weyl_polynomial_quantization('q0^2 p0^3 q1^3')

    where 'q' or 'p' is the phase space quadrature variable, the integer
    directly following is the mode it is with respect to, and '^2' is the
    polynomial power.

    Args:
        polynomial (str): polynomial function of q and p of the form
            'qi^m pj^n ...' where i,j are the modes, and m, n the powers.

    Returns:
        QuadOperator: the Weyl quantization of the phase space function.

    Warning:
        The runtime of this method is exponential in the maximum locality
        of the original operator.
    """
    # construct the equivalent QuadOperator
    poly = dict()

    if polynomial:
        for term in polynomial.split():
            if '^' in term:
                op, pwr = term.split('^')
                pwr = int(pwr)
            else:
                op = term
                pwr = 1

            mode = int(op[1:])

            if mode not in poly:
                poly[mode] = [0, 0]

            if op[0] == 'q':
                poly[mode][0] += pwr
            elif op[0] == 'p':
                poly[mode][1] += pwr

        # replace {q_i^m p_i^n} -> S({q_i^m p_i^n})
        operator = QuadOperator('')
        for mode, (m, n) in poly.items():
            qtmp = QuadOperator()
            qtmp.terms = mccoy(mode, 'q', 'p', m, n)
            operator *= qtmp
    else:
        operator = QuadOperator.zero()

    return operator


def symmetric_ordering(operator, ignore_coeff=True, ignore_identity=True):
    """ Apply the symmetric ordering to a BosonOperator or QuadOperator.

    The symmetric ordering is performed by applying McCoy's formula
    directly to polynomial terms of quadrature operators:

    q^m p^n -> (1/ 2^n) sum_{r=0}^{n} Binomial(n, r) q^r p^m q^{n-r}

    Note: in general, symmetric ordering is performed on a single term
    containing the tensor product of various operators. However, this
    function can also be applied to a sum of these terms, and the symmetric
    product is distributed over the summed terms.

    In this case, Hermiticity cannot be guaranteed - as such, by default
    term coefficients and identity operators are ignored. However, this
    behavior can be modified via keyword arguments describe below if necessary.

    Args:
        operator: either a BosonOperator or QuadOperator.
        ignore_coeff (bool): By default, the coefficients for
            each term are ignored; S(a q^m p^n) = S(q^m p^n), and
            the returned operator is always Hermitian.
            If set to False, then instead the coefficients are taken into
            account; S(q^m p^n) = a S(q^m p^n). In this case, if
            a is a complex coefficient, it is not guaranteed that the
            the returned operator will be Hermitian.
        ignore_identity (bool): By default, identity terms are ignore;
            S(I) = 0. If set to False, then instead S(I) = I.

    Returns:
        transformed_operator: an operator of the same class as in the input.

    Warning:
        The runtime of this method is exponential in the maximum locality
        of the original operator.
    """
    if isinstance(operator, BosonOperator):
        transformed_operator = BosonOperator()
        for term in operator.terms:
            if ignore_coeff:
                coeff = 1
            else:
                coeff = operator.terms[term]

            # Initialize identity matrix.
            transformed_term = BosonOperator('', coeff)

            if term:
                # convert term into the form \prod_i {bd_i^m b_i^n}
                modes = dict()
                for op in term:
                    if op[0] not in modes:
                        modes[op[0]] = [0, 0]

                    modes[op[0]][1 - op[1]] += 1

                # Replace {bd_i^m b_i^n} -> S({bd_i^m b_i^n})
                for mode, (m, n) in modes.items():
                    qtmp = BosonOperator()
                    qtmp.terms = mccoy(mode, 1, 0, m, n)
                    transformed_term *= qtmp

            if term or (not ignore_identity):
                transformed_operator += transformed_term

    elif isinstance(operator, QuadOperator):
        transformed_operator = QuadOperator()
        for term in operator.terms:
            if ignore_coeff:
                coeff = 1
            else:
                coeff = operator.terms[term]

            # Initialize identity matrix.
            transformed_term = QuadOperator('', coeff)

            if term:
                # convert term into the form \prod_i {q_i^m p_i^n}
                modes = dict()
                for op in term:
                    if op[0] not in modes:
                        modes[op[0]] = [0, 0]

                    if op[1] == 'q':
                        modes[op[0]][0] += 1
                    elif op[1] == 'p':
                        modes[op[0]][1] += 1

                # replace {q_i^m p_i^n} -> S({q_i^m p_i^n})
                for mode, (m, n) in modes.items():
                    qtmp = QuadOperator()
                    qtmp.terms = mccoy(mode, 'q', 'p', m, n)
                    transformed_term *= qtmp

            if term or (not ignore_identity):
                transformed_operator += transformed_term

    else:
        raise TypeError("operator must be a BosonOperator or " "QuadOperator.")

    return transformed_operator



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for multiband Hubbard model module."""

from collections import defaultdict
import itertools
import random
import pytest

from openfermion.hamiltonians import fermi_hubbard, FermiHubbardModel
from openfermion.utils import HubbardSquareLattice, SpinPairs
from openfermion.hamiltonians.general_hubbard import (InteractionParameter,
                                                      number_operator)

from openfermion.ops import FermionOperator


def fermi_hubbard_from_general(x_dimension,
                               y_dimension,
                               tunneling,
                               coulomb,
                               chemical_potential=0.,
                               periodic=True,
                               spinless=False,
                               magnetic_field=0):
    lattice = HubbardSquareLattice(x_dimension,
                                   y_dimension,
                                   periodic=periodic,
                                   spinless=spinless)
    interaction_edge_type = 'neighbor' if spinless else 'onsite'
    model = FermiHubbardModel(lattice,
                              tunneling_parameters=(('neighbor', (0, 0),
                                                     tunneling),),
                              interaction_parameters=((interaction_edge_type,
                                                       (0, 0), coulomb),),
                              potential_parameters=((0, chemical_potential),),
                              magnetic_field=magnetic_field)
    return model.hamiltonian()


@pytest.mark.parametrize(
    'x_dimension,y_dimension,tunneling,coulomb,' +
    'chemical_potential,spinless,periodic,magnetic_field',
    itertools.product(range(1, 4), range(1, 4), (random.uniform(0, 2.),),
                      (random.uniform(0, 2.),), (random.uniform(0, 2.),),
                      (True, False), (True, False), (random.uniform(-1, 1),)))
def test_fermi_hubbard_square_special_general_equivalence(
        x_dimension, y_dimension, tunneling, coulomb, chemical_potential,
        spinless, periodic, magnetic_field):
    hubbard_model_special = fermi_hubbard(x_dimension,
                                          y_dimension,
                                          tunneling,
                                          coulomb,
                                          chemical_potential=chemical_potential,
                                          spinless=spinless,
                                          periodic=periodic,
                                          magnetic_field=magnetic_field)
    hubbard_model_general = fermi_hubbard_from_general(
        x_dimension,
        y_dimension,
        tunneling,
        coulomb,
        chemical_potential=chemical_potential,
        spinless=spinless,
        periodic=periodic,
        magnetic_field=magnetic_field)
    assert hubbard_model_special == hubbard_model_general


def random_parameters(lattice, probability=0.5, distinguish_edges=False):
    parameters = {}
    edge_types = (('onsite', 'horizontal_neighbor',
                   'vertical_neighbor') if distinguish_edges else
                  ('onsite', 'neighbor'))

    parameters['tunneling_parameters'] = [
        (edge_type, dofs, random.uniform(-1, 1))
        for edge_type in edge_types
        for dofs in lattice.dof_pairs_iter(edge_type == 'onsite')
        if random.random() <= probability
    ]

    possible_spin_pairs = (SpinPairs.ALL,) if lattice.spinless else (
        SpinPairs.SAME, SpinPairs.DIFF)
    parameters['interaction_parameters'] = [
        (edge_type, dofs, random.uniform(-1, 1), spin_pairs)
        for edge_type in edge_types for spin_pairs in possible_spin_pairs
        for dofs in lattice.dof_pairs_iter(edge_type == 'onsite' and spin_pairs
                                           in (SpinPairs.ALL, SpinPairs.SAME))
        if random.random() <= probability
    ]

    parameters['potential_parameters'] = [(dof, random.uniform(-1, 1))
                                          for dof in lattice.dof_indices
                                          if random.random() <= probability]

    if random.random() <= probability:
        parameters['magnetic_field'] = random.uniform(-1, 1)

    return parameters


def test_fermi_hubbard_default_parameters():
    lattice = HubbardSquareLattice(3, 3)
    model = FermiHubbardModel(lattice)
    assert model.tunneling_parameters == []
    assert model.interaction_parameters == []
    assert model.potential_parameters == []
    assert model.magnetic_field == 0


def test_fermi_hubbard_bad_parameters():
    lattice = HubbardSquareLattice(3, 3)
    with pytest.raises(ValueError):
        tunneling_parameters = [('onsite', (0, 0), 1)]
        FermiHubbardModel(lattice, tunneling_parameters=tunneling_parameters)

    with pytest.raises(ValueError):
        FermiHubbardModel(lattice, interaction_parameters=[(0, 0)])
    with pytest.raises(ValueError):
        FermiHubbardModel(lattice, interaction_parameters=[(0,) * 5])
    with pytest.raises(ValueError):
        interaction_parameters = [('onsite', (0, 0), 1, SpinPairs.SAME)]
        FermiHubbardModel(lattice,
                          interaction_parameters=interaction_parameters)


lattices = [
    HubbardSquareLattice(random.randrange(3, 10),
                         random.randrange(3, 10),
                         periodic=periodic,
                         spinless=spinless) for periodic in (False, True)
    for spinless in (False, True) for _ in range(2)
]


@pytest.mark.parametrize(
    'lattice,parameters,distinguish_edges',
    [(lattice, random_parameters(
        lattice, distinguish_edges=distinguish_edges), distinguish_edges)
     for lattice in lattices for _ in range(2)
     for distinguish_edges in (True, False)])
def test_fermi_hubbard_square_lattice_random_parameters(lattice, parameters,
                                                        distinguish_edges):
    model = FermiHubbardModel(lattice, **parameters)
    hamiltonian = model.hamiltonian()
    terms_per_parameter = defaultdict(int)
    for term, coefficient in hamiltonian.terms.items():
        spin_orbitals = set(i for i, _ in term)
        if len(spin_orbitals) == 2:
            (i, a, s), (ii, aa, ss) = (
                lattice.from_spin_orbital_index(i) for i in spin_orbitals)
            edge_type = ({
                (0, 0): 'onsite',
                (0, 1): 'vertical_neighbor',
                (1, 0): 'horizontal_neighbor'
            } if distinguish_edges else {
                (0, 0): 'onsite',
                (0, 1): 'neighbor',
                (1, 0): 'neighbor'
            })[lattice.delta_mag(i, ii, True)]
            dofs = tuple(sorted((a, aa)))
            if len(term) == 2:
                parameter = (edge_type, dofs, -coefficient)
                assert parameter in parameters['tunneling_parameters']
                terms_per_parameter['tunneling', parameter] += 1
            else:
                assert len(term) == 4
                spin_pairs = (SpinPairs.ALL if lattice.spinless else
                              (SpinPairs.SAME if s == ss else SpinPairs.DIFF))
                parameter = (edge_type, dofs, coefficient, spin_pairs)
                assert parameter in parameters['interaction_parameters']
                terms_per_parameter['interaction', parameter] += 1
        else:
            assert len(term) == 2
            assert len(spin_orbitals) == 1
            spin_orbital = spin_orbitals.pop()
            _, dof, spin_index = lattice.from_spin_orbital_index(spin_orbital)
            potential_coefficient = -coefficient
            if not lattice.spinless:
                (-1)**spin_index
                potential_coefficient -= (((-1)**spin_index) *
                                          parameters.get('magnetic_field', 0))
            if not potential_coefficient:
                continue
            parameter = (dof, potential_coefficient)
            assert parameter in parameters['potential_parameters']
            terms_per_parameter['potential', parameter] += 1
    edge_type_to_n_site_pairs = {
        'onsite': lattice.n_sites,
        'neighbor': lattice.n_neighbor_pairs(False),
        'vertical_neighbor': lattice.n_vertical_neighbor_pairs(False),
        'horizontal_neighbor': lattice.n_horizontal_neighbor_pairs(False)
    }
    for (term_type, parameter), n_terms in terms_per_parameter.items():
        if term_type == 'potential':
            assert n_terms == lattice.n_sites * lattice.n_spin_values
            continue
        n_site_pairs = edge_type_to_n_site_pairs[parameter[0]]
        if term_type == 'tunneling':
            assert n_terms == 2 * n_site_pairs * lattice.n_spin_values
        else:
            assert term_type == 'interaction'
            parameter = InteractionParameter(*parameter)
            expected_n_terms = n_site_pairs
            if parameter.edge_type == 'neighbor':
                expected_n_terms *= len(set(parameter.dofs))
            if not lattice.spinless:
                assert parameter.spin_pairs != SpinPairs.ALL
                if (parameter.edge_type == 'onsite' and
                        parameter.spin_pairs == SpinPairs.DIFF):
                    expected_n_terms *= len(set(parameter.dofs))
                else:
                    expected_n_terms *= 2
            assert n_terms == expected_n_terms


def test_number_op():
    nop = FermionOperator(((0, 1), (0, 0)), coefficient=1.)
    test_op = number_operator(0)
    assert test_op == nop

    test_op = number_operator(0, particle_hole_symmetry=True)
    assert test_op == nop - FermionOperator((), coefficient=0.5)


#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import unittest
import numpy

from openfermion.utils import Grid
from openfermion.utils.grid import OrbitalSpecificationError
from openfermion.testing.testing_utils import EqualsTester


class GridTest(unittest.TestCase):

    def test_orbital_id(self):

        # Test in 1D with spin.
        grid = Grid(dimensions=1, length=5, scale=1.0)
        input_coords = [0, 1, 2, 3, 4]
        tensor_factors_up = [1, 3, 5, 7, 9]
        tensor_factors_down = [0, 2, 4, 6, 8]

        test_output_up = [grid.orbital_id(i, 1) for i in input_coords]
        test_output_down = [grid.orbital_id(i, 0) for i in input_coords]

        self.assertEqual(test_output_up, tensor_factors_up)
        self.assertEqual(test_output_down, tensor_factors_down)

        with self.assertRaises(OrbitalSpecificationError):
            grid.orbital_id(6, 1)

        # Test in 2D without spin.
        grid = Grid(dimensions=2, length=3, scale=1.0)
        input_coords = [(0, 0), (0, 1), (1, 2)]
        tensor_factors = [0, 3, 7]
        test_output = [grid.orbital_id(i) for i in input_coords]
        self.assertEqual(test_output, tensor_factors)

    def test_position_vector(self):

        # Test in 1D.
        grid = Grid(dimensions=1, length=4, scale=4.)
        test_output = [
            grid.position_vector(i)[0] for i in range(grid.length[0])
        ]
        correct_output = [-2, -1, 0, 1]
        self.assertEqual(correct_output, test_output)

        # Test in 2D.
        grid = Grid(dimensions=2, length=3, scale=3.)
        test_input = []
        test_output = []
        for i in range(3):
            for j in range(3):
                test_input += [(i, j)]
                test_output += [grid.position_vector((i, j))]
        correct_output = numpy.array([[-1., -1.], [-1., 0.], [-1., 1.],
                                      [0., -1.], [0., 0.], [0., 1.], [1., -1.],
                                      [1., 0.], [1., 1.]])
        self.assertAlmostEqual(0., numpy.amax(test_output - correct_output))

    def test_momentum_vector(self):
        grid = Grid(dimensions=1, length=3, scale=2. * numpy.pi)
        test_output = [grid.momentum_vector(i) for i in range(grid.length[0])]
        correct_output = [-1., 0, 1.]
        self.assertEqual(correct_output, test_output)

        grid = Grid(dimensions=1, length=2, scale=2. * numpy.pi)
        test_output = [grid.momentum_vector(i) for i in range(grid.length[0])]
        correct_output = [-1., 0.]
        self.assertEqual(correct_output, test_output)

        grid = Grid(dimensions=1, length=11, scale=2. * numpy.pi)
        for i in range(grid.length[0]):
            self.assertAlmostEqual(-grid.momentum_vector(i),
                                   grid.momentum_vector(grid.length[0] - i - 1))

        # Test in 2D.
        grid = Grid(dimensions=2, length=3, scale=2. * numpy.pi)
        test_input = []
        test_output = []
        for i in range(3):
            for j in range(3):
                test_input += [(i, j)]
                test_output += [grid.momentum_vector((i, j))]
        correct_output = numpy.array([[-1, -1], [-1, 0], [-1, 1], [0, -1],
                                      [0, 0], [0, 1], [1, -1], [1, 0], [1, 1]])
        self.assertAlmostEqual(0., numpy.amax(test_output - correct_output))

    def test_grid_indices(self):
        g1 = Grid(dimensions=2, length=4, scale=1.0)
        self.assertListEqual(g1.grid_indices(0, spinless=True), [0, 0])

        for elmt in g1.grid_indices(0, spinless=True):
            self.assertIsInstance(elmt, int)

        # Check that out-of-range qubit IDs raise correct errors.
        with self.assertRaises(OrbitalSpecificationError):
            _ = g1.grid_indices(-1, spinless=True)

        with self.assertRaises(OrbitalSpecificationError):
            _ = g1.grid_indices(16, spinless=True)

        with self.assertRaises(OrbitalSpecificationError):
            _ = g1.grid_indices(17, spinless=True)

        # Check that the spinful grid allows appropriate qubit IDs.
        self.assertListEqual(g1.grid_indices(31, spinless=False), [3, 3])

        # Check correct ordering of the list output.
        self.assertListEqual(g1.grid_indices(17, spinless=False), [0, 2])
        self.assertListEqual(g1.grid_indices(8, spinless=True), [0, 2])

    def test_initialize(self):
        g1 = Grid(dimensions=3, length=3, scale=1.0)
        scale_matrix = numpy.diag([1.0] * 3)
        g2 = Grid(dimensions=3, length=(3, 3, 3), scale=scale_matrix)
        self.assertEqual(g1, g2)

    def test_no_errors_in_call(self):
        # No exception
        _ = Grid(dimensions=1, length=1, scale=1.0)
        _ = Grid(dimensions=2, length=3, scale=0.01)
        _ = Grid(dimensions=23, length=34, scale=45.0)

    def test_preconditions_raise_value_error(self):
        nan = float('nan')

        with self.assertRaises(ValueError):
            _ = Grid(dimensions=0, length=0, scale=1.0)
        with self.assertRaises(ValueError):
            _ = Grid(dimensions=1, length=1, scale=1)
        with self.assertRaises(ValueError):
            _ = Grid(dimensions=1, length=1, scale=0.0)
        with self.assertRaises(ValueError):
            _ = Grid(dimensions=1, length=1, scale=-1.0)
        with self.assertRaises(ValueError):
            _ = Grid(dimensions=1, length=1, scale=nan)

        with self.assertRaises(ValueError):
            _ = Grid(dimensions=1, length=-1, scale=1.0)
        with self.assertRaises(ValueError):
            _ = Grid(dimensions=-1, length=1, scale=1.0)

    def test_position_and_momentum_vector_orbital_specification_error(self):
        with self.assertRaises(OrbitalSpecificationError):
            g = Grid(dimensions=1, length=2, scale=1.0)
            _ = g.position_vector((10,))
        with self.assertRaises(OrbitalSpecificationError):
            g = Grid(dimensions=1, length=2, scale=1.0)
            _ = g.momentum_vector((10,))

    def test_properties(self):
        g = Grid(dimensions=2, length=3, scale=5.0)
        self.assertEqual(g.num_points, 9)
        self.assertEqual(g.volume_scale(), 25)
        self.assertEqual(list(g.all_points_indices()), [
            (0, 0),
            (0, 1),
            (0, 2),
            (1, 0),
            (1, 1),
            (1, 2),
            (2, 0),
            (2, 1),
            (2, 2),
        ])

    def test_equality(self):
        eq = EqualsTester(self)
        eq.make_equality_pair(lambda: Grid(dimensions=5, length=3, scale=0.5))
        eq.add_equality_group(Grid(dimensions=4, length=3, scale=0.5))
        eq.add_equality_group(Grid(dimensions=5, length=4, scale=0.5))
        eq.add_equality_group(Grid(dimensions=5, length=3, scale=0.25))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Gates useful for simulating fermions."""

from .common_gates import (
    FSWAP,
    FSwapPowGate,
    Rxxyy,
    Ryxxy,
    Rzz,
    rot11,
)

from .fermionic_simulation import (
    state_swap_eigen_component,
    fermionic_simulation_gates_from_interaction_operator,
    sum_of_interaction_operator_gate_generators,
    ParityPreservingFermionicGate,
    InteractionOperatorFermionicGate,
    QuadraticFermionicSimulationGate,
    CubicFermionicSimulationGate,
    QuarticFermionicSimulationGate,
)

from .four_qubit_gates import (
    DoubleExcitation,
    DoubleExcitationGate,
)

from .three_qubit_gates import (
    rot111,
    CRxxyy,
    CRyxxy,
)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
OpenFermion

For more information, examples, or tutorials visit our website:

www.openfermion.org
"""
from openfermion.chem import (
    make_atom,
    make_atomic_lattice,
    make_atomic_ring,
    angstroms_to_bohr,
    bohr_to_angstroms,
    MolecularData,
    name_molecule,
    geometry_from_file,
    load_molecular_hamiltonian,
    periodic_table,
    periodic_hash_table,
    periodic_polarization,
    geometry_from_pubchem,
    make_reduced_hamiltonian,
)

from openfermion.functionals import contextuality

from openfermion.hamiltonians import (
    FermiHubbardModel,
    number_operator,
    interaction_operator,
    tunneling_operator,
    number_difference_operator,
    rhf_func_generator,
    rhf_minimization,
    HartreeFockFunctional,
    rhf_params_to_matrix,
    get_matrix_of_eigs,
    generate_hamiltonian,
    bose_hubbard,
    fermi_hubbard,
    dual_basis_kinetic,
    dual_basis_potential,
    dual_basis_jellium_model,
    jellium_model,
    jordan_wigner_dual_basis_jellium,
    hypercube_grid_with_given_wigner_seitz_radius_and_filling,
    plane_wave_kinetic,
    plane_wave_potential,
    wigner_seitz_length_scale,
    hartree_fock_state_jellium,
    mean_field_dwave,
    dual_basis_external_potential,
    plane_wave_external_potential,
    plane_wave_hamiltonian,
    jordan_wigner_dual_basis_hamiltonian,
    s_plus_operator,
    s_squared_operator,
    sx_operator,
    sy_operator,
    sz_operator,
    majorana_operator,
    number_operator,
)

from openfermion.linalg import (
    Davidson,
    DavidsonOptions,
    DavidsonError,
    QubitDavidson,
    SparseDavidson,
    generate_random_vectors,
    append_random_vectors,
    orthonormalize,
    erpa_eom_hamiltonian,
    singlet_erpa,
    givens_decomposition,
    givens_rotate,
    double_givens_rotate,
    givens_decomposition_square,
    givens_matrix_elements,
    fermionic_gaussian_decomposition,
    swap_rows,
    swap_columns,
    generate_linear_qubit_operator,
    LinearQubitOperator,
    LinearQubitOperatorOptions,
    ParallelLinearQubitOperator,
    wrapped_kronecker,
    kronecker_operators,
    jordan_wigner_ladder_sparse,
    jordan_wigner_sparse,
    qubit_operator_sparse,
    eigenspectrum,
    get_linear_qubit_operator_diagonal,
    jw_configuration_state,
    jw_hartree_fock_state,
    jw_number_indices,
    jw_sz_indices,
    jw_number_restrict_operator,
    jw_sz_restrict_operator,
    jw_number_restrict_state,
    jw_sz_restrict_state,
    jw_get_ground_state_at_particle_number,
    jw_sparse_givens_rotation,
    jw_sparse_particle_hole_transformation_last_mode,
    get_density_matrix,
    get_ground_state,
    sparse_eigenspectrum,
    expectation,
    variance,
    expectation_computational_basis_state,
    expectation_db_operator_with_pw_basis_state,
    expectation_one_body_db_operator_computational_basis_state,
    expectation_two_body_db_operator_computational_basis_state,
    expectation_three_body_db_operator_computational_basis_state,
    get_gap,
    inner_product,
    boson_ladder_sparse,
    single_quad_op_sparse,
    boson_operator_sparse,
    get_sparse_operator,
    get_number_preserving_sparse_operator,
    generate_parity_permutations,
    fit_known_frequencies,
    prony,
    wedge,
    valdemoro_reconstruction,
)

from openfermion.measurements import (
    pair_within,
    pair_between,
    pair_within_simultaneously,
    pair_within_simultaneously_binned,
    pair_within_simultaneously_symmetric,
    apply_constraints,
    constraint_matrix,
    linearize_term,
    unlinearize_term,
    operator_to_vector,
    vector_to_operator,
    get_interaction_rdm,
    group_into_tensor_product_basis_sets,
    one_body_fermion_constraints,
    two_body_fermion_constraints,
    binary_partition_iterator,
    partition_iterator,
    pauli_string_iterator,
    PhaseFitEstimator,
    get_phase_function,
)

from openfermion.ops import (
    BinaryPolynomial,
    BinaryPolynomialError,
    BosonOperator,
    FermionOperator,
    IsingOperator,
    MajoranaOperator,
    QuadOperator,
    QubitOperator,
    SymbolicOperator,
    double_decoding,
    shift_decoder,
    BinaryCode,
    BinaryCodeError,
    PolynomialTensor,
    PolynomialTensorError,
    general_basis_change,
    DiagonalCoulombHamiltonian,
    InteractionOperator,
    InteractionOperatorError,
    InteractionRDM,
    InteractionRDMError,
    QuadraticHamiltonian,
    QuadraticHamiltonianError,
    DOCIHamiltonian,
    antisymmetric_canonical_form,
)

# Shifted here to fix circular dependencies
from openfermion.circuits import (
    prepare_slater_determinant,
    FSWAP,
    FSwapPowGate,
    Rxxyy,
    Ryxxy,
    Rzz,
    rot11,
    state_swap_eigen_component,
    fermionic_simulation_gates_from_interaction_operator,
    sum_of_interaction_operator_gate_generators,
    ParityPreservingFermionicGate,
    InteractionOperatorFermionicGate,
    QuadraticFermionicSimulationGate,
    CubicFermionicSimulationGate,
    QuarticFermionicSimulationGate,
    DoubleExcitation,
    DoubleExcitationGate,
    rot111,
    CRxxyy,
    CRyxxy,
    preprocess_lcu_coefficients_for_reversible_sampling,
    lambda_norm,
    get_chemist_two_body_coefficients,
    low_rank_two_body_decomposition,
    prepare_one_body_squared_evolution,
    bogoliubov_transform,
    ffft,
    optimal_givens_decomposition,
    prepare_gaussian_state,
    swap_network,
    gaussian_state_preparation_circuit,
    slater_determinant_preparation_circuit,
    jw_get_gaussian_state,
    jw_slater_determinant,
    trotter_operator_grouping,
    pauli_exp_to_qasm,
    trotterize_exp_qubop_to_qasm,
    uccsd_generator,
    uccsd_convert_amplitude_format,
    uccsd_singlet_paramsize,
    uccsd_singlet_get_packed_amplitudes,
    uccsd_singlet_generator,
    LINEAR_SWAP_NETWORK,
    LinearSwapNetworkTrotterAlgorithm,
    LOW_RANK,
    LowRankTrotterAlgorithm,
    SplitOperatorTrotterAlgorithm,
    SplitOperatorTrotterStep,
    SymmetricSplitOperatorTrotterStep,
    ControlledAsymmetricSplitOperatorTrotterStep,
    AsymmetricSplitOperatorTrotterStep,
    ControlledSymmetricSplitOperatorTrotterStep,
    SPLIT_OPERATOR,
    diagonal_coulomb_potential_and_kinetic_terms_as_arrays,
    bit_mask_of_modes_acted_on_by_fermionic_terms,
    split_operator_trotter_error_operator_diagonal_two_body,
    fermionic_swap_trotter_error_operator_diagonal_two_body,
    simulation_ordered_grouped_hubbard_terms_with_info,
    low_depth_second_order_trotter_error_operator,
    low_depth_second_order_trotter_error_bound,
    simulation_ordered_grouped_low_depth_terms_with_info,
    stagger_with_info,
    simulate_trotter,
    TrotterAlgorithm,
    TrotterStep,
    error_bound,
    error_operator,
    trotter_steps_required,
    vpe_single_circuit,
    vpe_circuits_single_timestep,
    standard_vpe_rotation_set,
)

from openfermion.testing import (
    validate_trotterized_evolution,
    random_interaction_operator_term,
    haar_random_vector,
    random_antisymmetric_matrix,
    random_hermitian_matrix,
    random_unitary_matrix,
    random_qubit_operator,
    random_diagonal_coulomb_hamiltonian,
    random_interaction_operator,
    random_quadratic_hamiltonian,
    EqualsTester,
    module_importable,
    assert_equivalent_repr,
    assert_implements_consistent_protocols,
)

from openfermion.third_party import (
    fixed_trace_positive_projection,
    map_to_tensor,
    map_to_matrix,
    higham_polynomial,
    higham_root,
    heaviside,
)

from openfermion.transforms import (
    chemist_ordered,
    normal_ordered,
    normal_ordered_ladder_term,
    normal_ordered_quad_term,
    reorder,
    linearize_decoder,
    checksum_code,
    bravyi_kitaev_code,
    jordan_wigner_code,
    parity_code,
    weight_one_binary_addressing_code,
    weight_one_segment_code,
    weight_two_segment_code,
    interleaved_code,
    binary_code_transform,
    bravyi_kitaev_fast,
    bravyi_kitaev_fast_interaction_op,
    bravyi_kitaev_fast_edge_matrix,
    bravyi_kitaev,
    bravyi_kitaev_tree,
    get_fermion_operator,
    get_boson_operator,
    get_majorana_operator,
    get_quad_operator,
    check_no_sympy,
    FenwickNode,
    FenwickTree,
    jordan_wigner,
    jordan_wigner_one_body,
    jordan_wigner_two_body,
    qubit_operator_to_pauli_sum,
    reverse_jordan_wigner,
    symmetry_conserving_bravyi_kitaev,
    verstraete_cirac_2d_square,
    vertical_edges_snake,
    get_interaction_operator,
    get_diagonal_coulomb_hamiltonian,
    get_molecular_data,
    get_quadratic_hamiltonian,
    fourier_transform,
    inverse_fourier_transform,
    freeze_orbitals,
    prune_unused_indices,
    project_onto_sector,
    projection_error,
    rotate_qubit_by_pauli,
    StabilizerError,
    check_commuting_stabilizers,
    check_stabilizer_linearity,
    reduce_number_of_terms,
    taper_off_qubits,
    fix_single_term,
    mccoy,
    weyl_polynomial_quantization,
    symmetric_ordering,
)

from openfermion.utils import (
    bch_expand,
    amplitude_damping_channel,
    dephasing_channel,
    depolarizing_channel,
    HubbardSquareLattice,
    HubbardLattice,
    SpinPairs,
    Spin,
    anticommutator,
    commutator,
    double_commutator,
    trivially_double_commutes_dual_basis_using_term_info,
    trivially_commutes_dual_basis,
    trivially_double_commutes_dual_basis,
    Grid,
    up_index,
    down_index,
    up_then_down,
    count_qubits,
    get_file_path,
    hermitian_conjugated,
    is_hermitian,
    is_identity,
    load_operator,
    save_operator,
    OperatorUtilsError,
    OperatorSpecificationError,
    kronecker_delta,
    map_two_pdm_to_two_hole_dm,
    map_two_pdm_to_one_pdm,
    map_one_pdm_to_one_hole_dm,
    map_one_hole_dm_to_one_pdm,
    map_two_pdm_to_particle_hole_dm,
    map_two_hole_dm_to_two_pdm,
    map_two_hole_dm_to_one_hole_dm,
    map_particle_hole_dm_to_one_pdm,
    map_particle_hole_dm_to_two_pdm,
)

from ._version import __version__



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from itertools import product
import numpy
import scipy
import cirq

from openfermion.linalg import (givens_matrix_elements, givens_rotate,
                                get_sparse_operator)
from openfermion.ops import QubitOperator, FermionOperator
from openfermion.transforms import jordan_wigner

from openfermion.circuits.primitives.optimal_givens_decomposition import \
    optimal_givens_decomposition


def test_givens_inverse():
    r"""
    The Givens rotation in OpenFermion is defined as

    .. math::

        \begin{pmatrix}
            \cos(\theta) & -e^{i \varphi} \sin(\theta) \\
            \sin(\theta) &     e^{i \varphi} \cos(\theta)
        \end{pmatrix}.

    confirm numerically its hermitian conjugate is it's inverse
    """
    a = numpy.random.random() + 1j * numpy.random.random()
    b = numpy.random.random() + 1j * numpy.random.random()
    ab_rotation = givens_matrix_elements(a, b, which='right')

    assert numpy.allclose(ab_rotation.dot(numpy.conj(ab_rotation).T),
                          numpy.eye(2))
    assert numpy.allclose(
        numpy.conj(ab_rotation).T.dot(ab_rotation), numpy.eye(2))


def test_row_eliminate():
    """
    Test elemination of element in U[i, j] by rotating in i-1 and i.
    """
    dim = 3
    u_generator = numpy.random.random((dim, dim)) + 1j * numpy.random.random(
        (dim, dim))
    u_generator = u_generator - numpy.conj(u_generator).T

    # make sure the generator is actually antihermitian
    assert numpy.allclose(-1 * u_generator, numpy.conj(u_generator).T)

    unitary = scipy.linalg.expm(u_generator)

    # eliminate U[2, 0] by rotating in 1, 2
    gmat = givens_matrix_elements(unitary[1, 0], unitary[2, 0], which='right')
    givens_rotate(unitary, gmat, 1, 2, which='row')
    assert numpy.isclose(unitary[2, 0], 0.0)

    # eliminate U[1, 0] by rotating in 0, 1
    gmat = givens_matrix_elements(unitary[0, 0], unitary[1, 0], which='right')
    givens_rotate(unitary, gmat, 0, 1, which='row')
    assert numpy.isclose(unitary[1, 0], 0.0)

    # eliminate U[2, 1] by rotating in 1, 2
    gmat = givens_matrix_elements(unitary[1, 1], unitary[2, 1], which='right')
    givens_rotate(unitary, gmat, 1, 2, which='row')
    assert numpy.isclose(unitary[2, 1], 0.0)


def create_givens(givens_mat, i, j, dim):
    """
    Create the givens matrix on the larger space

    :param givens_mat: 2x2 matrix with first column is real
    :param i: row index i
    :param j: row index i < j
    :param dim: dimension
    """
    gmat = numpy.eye(dim, dtype=complex)
    gmat[i, i] = givens_mat[0, 0]
    gmat[i, j] = givens_mat[0, 1]
    gmat[j, i] = givens_mat[1, 0]
    gmat[j, j] = givens_mat[1, 1]
    return gmat


def test_col_eliminate():
    """
    Test elimination by rotating in the column space.  Left multiplication of
    inverse givens
    """
    dim = 3
    u_generator = numpy.random.random((dim, dim)) + 1j * numpy.random.random(
        (dim, dim))
    u_generator = u_generator - numpy.conj(u_generator).T
    # make sure the generator is actually antihermitian
    assert numpy.allclose(-1 * u_generator, numpy.conj(u_generator).T)
    unitary = scipy.linalg.expm(u_generator)

    # eliminate U[1, 0] by rotation in rows [0, 1] and
    # mixing U[1, 0] and U[0, 0]
    unitary_original = unitary.copy()
    gmat = givens_matrix_elements(unitary[0, 0], unitary[1, 0], which='right')
    vec = numpy.array([[unitary[0, 0]], [unitary[1, 0]]])
    fullgmat = create_givens(gmat, 0, 1, 3)
    zeroed_unitary = fullgmat.dot(unitary)

    givens_rotate(unitary, gmat, 0, 1)
    assert numpy.isclose(unitary[1, 0], 0.0)
    assert numpy.allclose(unitary.real, zeroed_unitary.real)
    assert numpy.allclose(unitary.imag, zeroed_unitary.imag)

    # eliminate U[2, 0] by rotating columns [0, 1] and
    # mixing U[2, 0] and U[2, 1].
    unitary = unitary_original.copy()
    gmat = givens_matrix_elements(unitary[2, 0], unitary[2, 1], which='left')
    vec = numpy.array([[unitary[2, 0]], [unitary[2, 1]]])

    assert numpy.isclose((gmat.dot(vec))[0, 0], 0.0)
    assert numpy.isclose((vec.T.dot(gmat.T))[0, 0], 0.0)
    fullgmat = create_givens(gmat, 0, 1, 3)
    zeroed_unitary = unitary.dot(fullgmat.T)

    # because col takes g[0, 0] * col_i + g[0, 1].conj() * col_j -> col_i
    # this is equivalent ot left multiplication by gmat.T
    givens_rotate(unitary, gmat.conj(), 0, 1, which='col')
    assert numpy.isclose(zeroed_unitary[2, 0], 0.0)
    assert numpy.allclose(unitary, zeroed_unitary)


def test_front_back_iteration():
    """
    Code demonstrating how we iterated over the matrix

    [[ 0.  0.  0.  0.  0.  0.]
     [15.  0.  0.  0.  0.  0.]
     [ 7. 14.  0.  0.  0.  0.]
     [ 6.  8. 13.  0.  0.  0.]
     [ 2.  5.  9. 12.  0.  0.]
     [ 1.  3.  4. 10. 11.  0.]]
    """
    N = 6
    unitary = numpy.zeros((N, N))
    unitary[-1, 0] = 1
    unitary[-2, 0] = 2
    unitary[-1, 1] = 3
    unitary[-1, 2] = 4
    unitary[-2, 1] = 5
    unitary[-3, 0] = 6
    unitary[-4, 0] = 7
    unitary[-3, 1] = 8
    unitary[-2, 2] = 9
    unitary[-1, 3] = 10
    unitary[-1, 4] = 11
    unitary[-2, 3] = 12
    unitary[-3, 2] = 13
    unitary[-4, 1] = 14
    unitary[-5, 0] = 15
    counter = 1
    for i in range(1, N):
        if i % 2 == 1:
            for j in range(0, i):
                print((N - j, i - j), i - j, i - j + 1, "col rotation")
                assert numpy.isclose(unitary[N - j - 1, i - j - 1], counter)
                counter += 1
        else:
            for j in range(1, i + 1):
                print((N + j - i, j), N + j - i - 1, N + j - i, "row rotation")
                assert numpy.isclose(unitary[N + j - i - 1, j - 1], counter)
                counter += 1


def test_circuit_generation_and_accuracy():
    for dim in range(2, 10):
        qubits = cirq.LineQubit.range(dim)
        u_generator = numpy.random.random(
            (dim, dim)) + 1j * numpy.random.random((dim, dim))
        u_generator = u_generator - numpy.conj(u_generator).T
        assert numpy.allclose(-1 * u_generator, numpy.conj(u_generator).T)

        unitary = scipy.linalg.expm(u_generator)
        circuit = cirq.Circuit()
        circuit.append(optimal_givens_decomposition(qubits, unitary))

        fermion_generator = QubitOperator(()) * 0.0
        for i, j in product(range(dim), repeat=2):
            fermion_generator += jordan_wigner(
                FermionOperator(((i, 1), (j, 0)), u_generator[i, j]))

        true_unitary = scipy.linalg.expm(
            get_sparse_operator(fermion_generator).toarray())
        assert numpy.allclose(true_unitary.conj().T.dot(true_unitary),
                              numpy.eye(2**dim, dtype=complex))

        test_unitary = cirq.unitary(circuit)
        assert numpy.isclose(
            abs(numpy.trace(true_unitary.conj().T.dot(test_unitary))), 2**dim)


def test_circuit_generation_state():
    """
    Determine if we rotate the Hartree-Fock state correctly
    """
    simulator = cirq.Simulator()
    circuit = cirq.Circuit()
    qubits = cirq.LineQubit.range(4)
    circuit.append([
        cirq.X(qubits[0]),
        cirq.X(qubits[1]),
        cirq.X(qubits[1]),
        cirq.X(qubits[2]),
        cirq.X(qubits[3]),
        cirq.X(qubits[3])
    ])  # alpha-spins are first then beta spins

    wavefunction = numpy.zeros((2**4, 1), dtype=complex)
    wavefunction[10, 0] = 1.0

    dim = 2
    u_generator = numpy.random.random((dim, dim)) + 1j * numpy.random.random(
        (dim, dim))
    u_generator = u_generator - numpy.conj(u_generator).T
    unitary = scipy.linalg.expm(u_generator)

    circuit.append(optimal_givens_decomposition(qubits[:2], unitary))

    fermion_generator = QubitOperator(()) * 0.0
    for i, j in product(range(dim), repeat=2):
        fermion_generator += jordan_wigner(
            FermionOperator(((i, 1), (j, 0)), u_generator[i, j]))

    test_unitary = scipy.linalg.expm(
        get_sparse_operator(fermion_generator, 4).toarray())
    test_final_state = test_unitary.dot(wavefunction)
    cirq_wf = simulator.simulate(circuit).final_state_vector

    assert numpy.allclose(cirq_wf, test_final_state.flatten())



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for slater_determinants.py."""
import unittest

import pytest

import numpy

from openfermion.linalg.sparse_tools import (
    jw_sparse_givens_rotation, jw_sparse_particle_hole_transformation_last_mode,
    get_sparse_operator, get_ground_state, jw_configuration_state)
from openfermion.testing.testing_utils import random_quadratic_hamiltonian

from openfermion.circuits.slater_determinants import (
    gaussian_state_preparation_circuit, jw_get_gaussian_state,
    jw_slater_determinant)


class JWSlaterDeterminantTest(unittest.TestCase):

    def test_hadamard_transform(self):
        r"""Test creating the states
        1 / sqrt(2) (a^\dagger_0 + a^\dagger_1) |vac>
        and
        1 / sqrt(2) (a^\dagger_0 - a^\dagger_1) |vac>.
        """
        slater_determinant_matrix = numpy.array([[1., 1.]]) / numpy.sqrt(2.)
        slater_determinant = jw_slater_determinant(slater_determinant_matrix)
        self.assertAlmostEqual(slater_determinant[1], slater_determinant[2])
        self.assertAlmostEqual(abs(slater_determinant[1]), 1. / numpy.sqrt(2.))
        self.assertAlmostEqual(abs(slater_determinant[0]), 0.)
        self.assertAlmostEqual(abs(slater_determinant[3]), 0.)

        slater_determinant_matrix = numpy.array([[1., -1.]]) / numpy.sqrt(2.)
        slater_determinant = jw_slater_determinant(slater_determinant_matrix)
        self.assertAlmostEqual(slater_determinant[1], -slater_determinant[2])
        self.assertAlmostEqual(abs(slater_determinant[1]), 1. / numpy.sqrt(2.))
        self.assertAlmostEqual(abs(slater_determinant[0]), 0.)
        self.assertAlmostEqual(abs(slater_determinant[3]), 0.)


class GaussianStatePreparationCircuitTest(unittest.TestCase):

    def setUp(self):
        self.n_qubits_range = range(3, 6)

    def test_ground_state_particle_conserving(self):
        """Test getting the ground state preparation circuit for a Hamiltonian
        that conserves particle number."""
        for n_qubits in self.n_qubits_range:
            print(n_qubits)
            # Initialize a particle-number-conserving Hamiltonian
            quadratic_hamiltonian = random_quadratic_hamiltonian(
                n_qubits, True, True)
            print(quadratic_hamiltonian)

            # Compute the true ground state
            sparse_operator = get_sparse_operator(quadratic_hamiltonian)
            ground_energy, _ = get_ground_state(sparse_operator)

            # Obtain the circuit
            circuit_description, start_orbitals = (
                gaussian_state_preparation_circuit(quadratic_hamiltonian))

            # Initialize the starting state
            state = jw_configuration_state(start_orbitals, n_qubits)

            # Apply the circuit
            for parallel_ops in circuit_description:
                for op in parallel_ops:
                    self.assertTrue(op != 'pht')
                    i, j, theta, phi = op
                    state = jw_sparse_givens_rotation(i, j, theta, phi,
                                                      n_qubits).dot(state)

            # Check that the state obtained using the circuit is a ground state
            difference = sparse_operator * state - ground_energy * state
            discrepancy = numpy.amax(numpy.abs(difference))
            self.assertAlmostEqual(discrepancy, 0)

    def test_ground_state_particle_nonconserving(self):
        """Test getting the ground state preparation circuit for a Hamiltonian
        that does not conserve particle number."""
        for n_qubits in self.n_qubits_range:
            # Initialize a particle-number-conserving Hamiltonian
            quadratic_hamiltonian = random_quadratic_hamiltonian(
                n_qubits, False, True)

            # Compute the true ground state
            sparse_operator = get_sparse_operator(quadratic_hamiltonian)
            ground_energy, _ = get_ground_state(sparse_operator)

            # Obtain the circuit
            circuit_description, start_orbitals = (
                gaussian_state_preparation_circuit(quadratic_hamiltonian))

            # Initialize the starting state
            state = jw_configuration_state(start_orbitals, n_qubits)

            # Apply the circuit
            particle_hole_transformation = (
                jw_sparse_particle_hole_transformation_last_mode(n_qubits))
            for parallel_ops in circuit_description:
                for op in parallel_ops:
                    if op == 'pht':
                        state = particle_hole_transformation.dot(state)
                    else:
                        i, j, theta, phi = op
                        state = jw_sparse_givens_rotation(
                            i, j, theta, phi, n_qubits).dot(state)

            # Check that the state obtained using the circuit is a ground state
            difference = sparse_operator * state - ground_energy * state
            discrepancy = numpy.amax(numpy.abs(difference))
            self.assertAlmostEqual(discrepancy, 0)

    def test_bad_input(self):
        """Test bad input."""
        with self.assertRaises(ValueError):
            gaussian_state_preparation_circuit('a')


class JWGetGaussianStateTest(unittest.TestCase):

    def setUp(self):
        self.n_qubits_range = range(2, 10)

    def test_ground_state_particle_conserving(self):
        """Test getting the ground state of a Hamiltonian that conserves
        particle number."""
        for n_qubits in self.n_qubits_range:
            # Initialize a particle-number-conserving Hamiltonian
            quadratic_hamiltonian = random_quadratic_hamiltonian(n_qubits, True)

            # Compute the true ground state
            sparse_operator = get_sparse_operator(quadratic_hamiltonian)
            ground_energy, _ = get_ground_state(sparse_operator)

            # Compute the ground state using the circuit
            circuit_energy, circuit_state = jw_get_gaussian_state(
                quadratic_hamiltonian)

            # Check that the energies match
            self.assertAlmostEqual(ground_energy, circuit_energy)

            # Check that the state obtained using the circuit is a ground state
            difference = (sparse_operator * circuit_state -
                          ground_energy * circuit_state)
            discrepancy = numpy.amax(numpy.abs(difference))
            self.assertAlmostEqual(discrepancy, 0)

    def test_ground_state_particle_nonconserving(self):
        """Test getting the ground state of a Hamiltonian that does not
        conserve particle number."""
        for n_qubits in self.n_qubits_range:
            # Initialize a non-particle-number-conserving Hamiltonian
            quadratic_hamiltonian = random_quadratic_hamiltonian(
                n_qubits, False)

            # Compute the true ground state
            sparse_operator = get_sparse_operator(quadratic_hamiltonian)
            ground_energy, _ = get_ground_state(sparse_operator)

            # Compute the ground state using the circuit
            circuit_energy, circuit_state = (
                jw_get_gaussian_state(quadratic_hamiltonian))

            # Check that the energies match
            self.assertAlmostEqual(ground_energy, circuit_energy)

            # Check that the state obtained using the circuit is a ground state
            difference = (sparse_operator * circuit_state -
                          ground_energy * circuit_state)
            discrepancy = numpy.amax(numpy.abs(difference))
            self.assertAlmostEqual(discrepancy, 0)

    def test_excited_state_particle_conserving(self):
        """Test getting an excited state of a Hamiltonian that conserves
        particle number."""
        for n_qubits in self.n_qubits_range:
            # Initialize a particle-number-conserving Hamiltonian
            quadratic_hamiltonian = random_quadratic_hamiltonian(n_qubits, True)

            # Pick some orbitals to occupy
            num_occupied_orbitals = numpy.random.randint(1, n_qubits + 1)
            occupied_orbitals = numpy.random.choice(range(n_qubits),
                                                    num_occupied_orbitals,
                                                    False)

            # Compute the Gaussian state
            circuit_energy, gaussian_state = jw_get_gaussian_state(
                quadratic_hamiltonian, occupied_orbitals)

            # Compute the true energy
            orbital_energies, constant = (
                quadratic_hamiltonian.orbital_energies())
            energy = numpy.sum(orbital_energies[occupied_orbitals]) + constant

            # Check that the energies match
            self.assertAlmostEqual(energy, circuit_energy)

            # Check that the state obtained using the circuit is an eigenstate
            # with the correct eigenvalue
            sparse_operator = get_sparse_operator(quadratic_hamiltonian)
            difference = (sparse_operator * gaussian_state -
                          energy * gaussian_state)
            discrepancy = numpy.amax(numpy.abs(difference))
            self.assertAlmostEqual(discrepancy, 0)

    def test_excited_state_particle_nonconserving(self):
        """Test getting an excited state of a Hamiltonian that conserves
        particle number."""
        for n_qubits in self.n_qubits_range:
            # Initialize a non-particle-number-conserving Hamiltonian
            quadratic_hamiltonian = random_quadratic_hamiltonian(
                n_qubits, False)

            # Pick some orbitals to occupy
            num_occupied_orbitals = numpy.random.randint(1, n_qubits + 1)
            occupied_orbitals = numpy.random.choice(range(n_qubits),
                                                    num_occupied_orbitals,
                                                    False)

            # Compute the Gaussian state
            circuit_energy, gaussian_state = jw_get_gaussian_state(
                quadratic_hamiltonian, occupied_orbitals)

            # Compute the true energy
            orbital_energies, constant = (
                quadratic_hamiltonian.orbital_energies())
            energy = numpy.sum(orbital_energies[occupied_orbitals]) + constant

            # Check that the energies match
            self.assertAlmostEqual(energy, circuit_energy)

            # Check that the state obtained using the circuit is an eigenstate
            # with the correct eigenvalue
            sparse_operator = get_sparse_operator(quadratic_hamiltonian)
            difference = (sparse_operator * gaussian_state -
                          energy * gaussian_state)
            discrepancy = numpy.amax(numpy.abs(difference))
            self.assertAlmostEqual(discrepancy, 0)

    def test_bad_input(self):
        """Test bad input."""
        with self.assertRaises(ValueError):
            jw_get_gaussian_state('a')


def test_not_implemented_spinr_reduced():
    """Tests that currently un-implemented functionality is caught."""
    msg = "Specifying spin sector for non-particle-conserving "
    msg += "Hamiltonians is not yet supported."
    for n_qubits in [2, 4, 6]:
        # Initialize a particle-number-conserving Hamiltonian
        quadratic_hamiltonian = random_quadratic_hamiltonian(
            n_qubits, False, True)

        # Obtain the circuit
        with pytest.raises(NotImplementedError):
            _ = gaussian_state_preparation_circuit(quadratic_hamiltonian,
                                                   spin_sector=1)

# if __name__ == "__main__":
#     inst = GaussianStatePreparationCircuitTest()
#     inst.setUp()
#     inst.test_ground_state_particle_conserving()


#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import unittest

from openfermion.testing.testing_utils import EqualsTester

from openfermion.ops.operators.ising_operator import IsingOperator


def test_properties():
    operator = IsingOperator()
    assert operator.actions == ('Z',)
    assert operator.action_strings == ('Z',)
    assert operator.action_before_index
    assert operator.different_indices_commute


class GeneralTest(unittest.TestCase):
    """General tests."""

    def test_ising_operator(self):
        equals_tester = EqualsTester(self)

        group = [IsingOperator('Z0 Z3'), IsingOperator([(0, 'Z'), (3, 'Z')])]

        equals_tester.add_equality_group(*group)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
'''Tests for the qubit_operator_transforms module'''

import unittest
import numpy

from openfermion.ops.operators import QubitOperator, FermionOperator
from openfermion.transforms.repconversions import (project_onto_sector,
                                                   projection_error,
                                                   rotate_qubit_by_pauli)
from openfermion.utils import count_qubits


class ProjectionTest(unittest.TestCase):

    def setUp(self):
        pass

    def test_function_errors(self):
        """Test main function errors."""
        operator = (QubitOperator('Z0 X1', 1.0) + QubitOperator('X1', 2.0))
        sector1 = [0]
        sector2 = [1]
        qbt_list = [0]
        with self.assertRaises(TypeError):
            project_onto_sector(operator=1.0, qubits=qbt_list, sectors=sector1)
        with self.assertRaises(TypeError):
            projection_error(operator=1.0, qubits=qbt_list, sectors=sector1)
        with self.assertRaises(TypeError):
            project_onto_sector(operator=operator, qubits=0.0, sectors=sector2)
        with self.assertRaises(TypeError):
            projection_error(operator=operator, qubits=0.0, sectors=sector2)
        with self.assertRaises(TypeError):
            project_onto_sector(operator=operator,
                                qubits=qbt_list,
                                sectors=operator)
        with self.assertRaises(TypeError):
            projection_error(operator=operator,
                             qubits=qbt_list,
                             sectors=operator)
        with self.assertRaises(ValueError):
            project_onto_sector(operator=operator,
                                qubits=[0, 1],
                                sectors=sector1)
        with self.assertRaises(ValueError):
            projection_error(operator=operator, qubits=[0, 1], sectors=sector1)
        with self.assertRaises(ValueError):
            project_onto_sector(operator=operator,
                                qubits=qbt_list,
                                sectors=[0, 0])
        with self.assertRaises(ValueError):
            projection_error(operator=operator, qubits=qbt_list, sectors=[0, 0])
        with self.assertRaises(ValueError):
            project_onto_sector(operator=operator,
                                qubits=qbt_list,
                                sectors=[-1])
        with self.assertRaises(ValueError):
            projection_error(operator=operator, qubits=qbt_list, sectors=[-1])

    def test_projection(self):
        coefficient = 0.5
        opstring = ((0, 'X'), (1, 'X'), (2, 'Z'))
        opstring2 = ((0, 'X'), (2, 'Z'), (3, 'Z'))
        operator = QubitOperator(opstring, coefficient)
        operator += QubitOperator(opstring2, coefficient)
        new_operator = project_onto_sector(operator,
                                           qubits=[2, 3],
                                           sectors=[0, 1])
        error = projection_error(operator, qubits=[2, 3], sectors=[0, 1])
        self.assertEqual(count_qubits(new_operator), 2)
        self.assertEqual(error, 0)
        self.assertTrue(((0, 'X'), (1, 'X')) in new_operator.terms)
        self.assertEqual(new_operator.terms[((0, 'X'), (1, 'X'))], 0.5)
        self.assertTrue(((0, 'X'),) in new_operator.terms)
        self.assertEqual(new_operator.terms[((0, 'X'),)], -0.5)

    def test_projection_error(self):
        coefficient = 0.5
        opstring = ((0, 'X'), (1, 'X'), (2, 'Z'))
        opstring2 = ((0, 'X'), (2, 'Z'), (3, 'Z'))
        operator = QubitOperator(opstring, coefficient)
        operator += QubitOperator(opstring2, coefficient)
        new_operator = project_onto_sector(operator, qubits=[1], sectors=[0])
        error = projection_error(operator, qubits=[1], sectors=[0])
        self.assertEqual(count_qubits(new_operator), 3)
        self.assertTrue(((0, 'X'), (1, 'Z'), (2, 'Z')) in new_operator.terms)
        self.assertEqual(new_operator.terms[((0, 'X'), (1, 'Z'), (2, 'Z'))],
                         0.5)
        self.assertEqual(error, 0.5)


class UnitaryRotationsTest(unittest.TestCase):

    def setup(self):
        pass

    def test_rotation(self):
        qop = QubitOperator('X0 X1', 1)
        qop += QubitOperator('Z0 Z1', 1)
        rot_op = QubitOperator('Z1', 1)

        rotated_qop = rotate_qubit_by_pauli(qop, rot_op, numpy.pi / 4)
        comp_op = QubitOperator('Z0 Z1', 1)
        comp_op += QubitOperator('X0 Y1', 1)
        self.assertEqual(comp_op, rotated_qop)

    def test_exception_Pauli(self):
        qop = QubitOperator('X0 X1', 1)
        qop += QubitOperator('Z0 Z1', 1)
        rot_op = QubitOperator('Z1', 1)
        rot_op += QubitOperator('Z0', 1)
        rot_op2 = QubitOperator('Z1', 1)
        ferm_op = FermionOperator('1^ 2', 1)
        with self.assertRaises(TypeError):
            rotate_qubit_by_pauli(qop, rot_op, numpy.pi / 4)
        with self.assertRaises(TypeError):
            rotate_qubit_by_pauli(ferm_op, rot_op2, numpy.pi / 4)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Perform Hamiltonian simulation via a Trotter-Suzuki product formula."""

from typing import Optional, Sequence

import cirq

from openfermion import ops
from openfermion.circuits.trotter.trotter_algorithm import (Hamiltonian,
                                                            TrotterStep,
                                                            TrotterAlgorithm)
from openfermion.circuits.trotter.algorithms import (LINEAR_SWAP_NETWORK,
                                                     LOW_RANK)


def simulate_trotter(qubits: Sequence[cirq.Qid],
                     hamiltonian: Hamiltonian,
                     time: float,
                     n_steps: int = 1,
                     order: int = 0,
                     algorithm: Optional[TrotterAlgorithm] = None,
                     control_qubit: Optional[cirq.Qid] = None,
                     omit_final_swaps: bool = False) -> cirq.OP_TREE:
    """Simulate Hamiltonian evolution using a Trotter-Suzuki product formula.

    The input is a Hamiltonian represented as an InteractionOperator or
    DiagonalCoulombHamiltonian. Not all types are supported by all algorithm
    options.

    The product formula used is from "General theory of fractal path integrals
    with applications to many-body theories and statistical physics" by
    Masuo Suzuki.

    Args:
        qubits: The qubits on which to apply operations. They should be sorted
            so that the j-th qubit in the Sequence holds the occupation of the
            j-th fermionic mode.
        hamiltonian: The Hamiltonian to simulate.
        time: The evolution time.
        n_steps: The number of Trotter steps to use. Default is 1.
        order: The order of the product formula. The value indexes symmetric
            formulae, e.g., a value of 2 indicates a second-order symmetric,
            sometimes known as a fourth-order, Trotter formula. A value of 0
            indicates an asymmetric Trotter formula. Default is 0.
        algorithm: The algorithm to use to simulate a single Trotter step.
            This is a constant exposed in the openfermion.trotter module.
            If not specified, a default option will be chosen based on the
            type of the given Hamiltonian.
            Available options:
                LINEAR_SWAP_NETWORK: The algorithm from arXiv:1711.04789.
                    Requires the input to be a DiagonalCoulombHamiltonian.
                LOW_RANK: The "low rank" strategy.
                    Requires the input to be an InteractionOperator.
                SPLIT_OPERATOR: The algorithm from arXiv:1706.00023.
                    Requires the input to be a DiagonalCoulombHamiltonian.
        control_qubit: A qubit on which to control the Trotter step.
        omit_final_swaps: If this is set to True, then SWAP or FSWAP gates at
            the end of the circuit may be omitted. This option exists because
            certain Trotter step algorithms, such as those based on swap
            networks, induce a permutation on the qubits or on the ordering in
            which qubits represent fermionic modes. For instance, algorithms
            based on swap networks may reverse the qubits depending on the
            number of Trotter steps used and the order of the Trotter formula
            selected. Setting this option to True will sometimes result in a
            circuit with fewer gates, but with the ordering of qubits or modes
            reversed in the final wavefunction.
    """
    # TODO Document gate complexities of algorithm options
    if order < 0:
        raise ValueError('The order of the Trotter formula must be at least 0.')

    if algorithm is None:
        algorithm = _select_trotter_algorithm(hamiltonian)

    if not isinstance(hamiltonian, tuple(algorithm.supported_types)):
        raise TypeError(
            'The input Hamiltonian was a {} but the chosen Trotter step '
            'algorithm only supports Hamiltonians of type {}'.format(
                type(hamiltonian).__name__,
                {cls.__name__ for cls in algorithm.supported_types}))

    # Select the Trotter step to use
    trotter_step = _select_trotter_step(hamiltonian,
                                        order,
                                        algorithm,
                                        controlled=control_qubit is not None)

    # Get ready to perform Trotter steps
    yield trotter_step.prepare(qubits, control_qubit)

    # Perform Trotter steps
    step_time = time / n_steps
    for _ in range(n_steps):
        yield _perform_trotter_step(qubits, step_time, order, trotter_step,
                                    control_qubit)
        qubits, control_qubit = trotter_step.step_qubit_permutation(
            qubits, control_qubit)

    # Finish
    yield trotter_step.finish(qubits, n_steps, control_qubit, omit_final_swaps)


def _perform_trotter_step(qubits: Sequence[cirq.Qid], time: float, order: int,
                          trotter_step: TrotterStep,
                          control_qubit: Optional[cirq.Qid]) -> cirq.OP_TREE:
    """Perform a Trotter step."""
    if order <= 1:
        yield trotter_step.trotter_step(qubits, time, control_qubit)
    else:
        # Recursively split this Trotter step into five smaller steps.
        # The first two and last two steps use this amount of time
        split_time = time / (4 - 4**(1 / (2 * order - 1)))

        for _ in range(2):
            yield _perform_trotter_step(qubits, split_time, order - 1,
                                        trotter_step, control_qubit)
            qubits, control_qubit = trotter_step.step_qubit_permutation(
                qubits, control_qubit)

        yield _perform_trotter_step(qubits, time - 4 * split_time, order - 1,
                                    trotter_step, control_qubit)
        qubits, control_qubit = trotter_step.step_qubit_permutation(
            qubits, control_qubit)

        for _ in range(2):
            yield _perform_trotter_step(qubits, split_time, order - 1,
                                        trotter_step, control_qubit)
            qubits, control_qubit = trotter_step.step_qubit_permutation(
                qubits, control_qubit)


def _select_trotter_algorithm(hamiltonian: Hamiltonian) -> TrotterAlgorithm:
    if isinstance(hamiltonian, ops.DiagonalCoulombHamiltonian):
        return LINEAR_SWAP_NETWORK
    elif isinstance(hamiltonian, ops.InteractionOperator):
        return LOW_RANK
    else:
        raise TypeError('Failed to select a default Trotter algorithm '
                        'for Hamiltonian of type {}.'.format(
                            type(hamiltonian).__name__))


def _select_trotter_step(hamiltonian: Hamiltonian, order: int,
                         algorithm: TrotterAlgorithm,
                         controlled: bool) -> TrotterStep:
    """Select a particular Trotter step from a Trotter step algorithm."""
    if controlled:
        if order == 0:
            trotter_step = algorithm.controlled_asymmetric(hamiltonian)
            if trotter_step is None:
                raise ValueError('The chosen Trotter step algorithm does not '
                                 'support the order 0 (asymmetric) formula '
                                 'with a control qubit.')
        else:
            trotter_step = algorithm.controlled_symmetric(hamiltonian)
            if trotter_step is None:
                raise ValueError('The chosen Trotter step algorithm does not '
                                 'support higher (> 0) order formulas '
                                 'with a control qubit.')
    else:
        if order == 0:
            trotter_step = algorithm.asymmetric(hamiltonian)
            if trotter_step is None:
                raise ValueError('The chosen Trotter step algorithm does not '
                                 'support the order 0 (asymmetric) formula.')
        else:
            trotter_step = algorithm.symmetric(hamiltonian)
            if trotter_step is None:
                raise ValueError('The chosen Trotter step algorithm does not '
                                 'support higher (> 0) order formulas.')
    return trotter_step



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for vpe_circuits.py"""

import numpy
import cirq

from openfermion.measurements import get_phase_function

from .vpe_circuits import (
    vpe_single_circuit,
    vpe_circuits_single_timestep,
)


def test_single_circuit():
    q0 = cirq.GridQubit(0, 0)
    q1 = cirq.GridQubit(0, 1)
    qubits = reversed([q0, q1])
    prep = cirq.Circuit([cirq.FSimGate(theta=numpy.pi / 4, phi=0).on(q0, q1)])
    evolve = cirq.Circuit(
        [cirq.rz(numpy.pi / 2).on(q0),
         cirq.rz(numpy.pi / 2).on(q1)])
    initial_rotation = cirq.ry(numpy.pi / 2).on(q0)
    final_rotation = cirq.rx(-numpy.pi / 2).on(q0)
    circuit = vpe_single_circuit(qubits, prep, evolve, initial_rotation,
                                 final_rotation)
    assert len(circuit) == 6
    simulator = cirq.Simulator()
    result = simulator.run(circuit, repetitions=100)
    data_counts = result.data['msmt'].value_counts()
    assert data_counts[1] == 100


def test_single_timestep():
    q0 = cirq.GridQubit(0, 0)
    q1 = cirq.GridQubit(0, 1)
    qubits = [q0, q1]
    prep = cirq.Circuit([cirq.FSimGate(theta=numpy.pi / 4, phi=0).on(q0, q1)])
    evolve = cirq.Circuit(
        [cirq.rz(numpy.pi / 2).on(q0),
         cirq.rz(numpy.pi / 2).on(q1)])
    target_qubit = q0
    circuits = vpe_circuits_single_timestep(qubits, prep, evolve, target_qubit)
    results = []
    simulator = cirq.Simulator()
    for circuit in circuits:
        this_res = simulator.run(circuit, repetitions=10000)
        results.append(this_res)
    pf_estimation = get_phase_function(results, qubits, 0)
    # Loose bound - standard deviation on this measurement should be 0.005, so
    # 95% of times this should pass. If it fails, check the absolute number.
    assert numpy.abs(pf_estimation - 1j) < 1e-2



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""The Bogoliubov transformation."""

from typing import (Iterable, List, Optional, Sequence, Tuple, Union, cast)

import numpy

import cirq

from openfermion import circuits, linalg


def bogoliubov_transform(
        qubits: Sequence[cirq.Qid],
        transformation_matrix: numpy.ndarray,
        initial_state: Optional[Union[int, Sequence[int]]] = None
) -> cirq.OP_TREE:
    r"""Perform a Bogoliubov transformation.

    This circuit performs the transformation to a basis determined by a new set
    of fermionic ladder operators. It performs the unitary :math:`U` such that

    .. math::

        U a^\dagger_p U^{-1} = b^\dagger_p

    where the :math:`a^\dagger_p` are the original creation operators and the
    :math:`b^\dagger_p` are the new creation operators. The new creation
    operators are linear combinations of the original ladder operators with
    coefficients given by the matrix `transformation_matrix`, which will be
    referred to as :math:`W` in the following.

    If :math:`W` is an `N \times N` matrix, then the :math:`b^\dagger_p` are
    given by

    .. math::

        b^\dagger_p = \sum_{q=1}^N W_{pq} a^\dagger_q.

    If :math:`W` is an `N \times 2N` matrix, then the :math:`b^\dagger_p` are
    given by

    .. math::

        b^\dagger_p = \sum_{q=1}^N W_{pq} a^\dagger_q
                      + \sum_{q=N+1}^{2N} W_{pq} a_q.

    This algorithm assumes the Jordan-Wigner Transform.

    Args:
        qubits: The qubits to which to apply the circuit.
        transformation_matrix: The matrix :math:`W` holding the coefficients
            that describe the new creation operators in terms of the original
            ladder operators. Its shape should be either :math:`NxN` or
            :math:`Nx(2N)`, where :math:`N` is the number of qubits.
        initial_state: Optionally specifies a computational basis state
            to assume that the qubits start in. This assumption enables
            optimizations that result in a circuit with fewer gates.
            This can be either an integer or a sequence of integers.
            If an integer, it is mapped to a computational basis state via
            "big endian" ordering of the binary representation of the integer.
            For example, the computational basis state on five qubits with
            the first and second qubits set to one is 0b11000, which is 24
            in decimal.
            If a sequence of integers, then it contains the indices of the
            qubits that are set to one (indexing starts from 0). For
            example, the list [2, 3] represents qubits 2 and 3 being set to one.
            Default is 0, the all zeros state.
    """

    n_qubits = len(qubits)
    shape = transformation_matrix.shape
    if shape not in [(n_qubits, n_qubits), (n_qubits, 2 * n_qubits)]:
        raise ValueError('Bad shape for transformation_matrix. '
                         'Expected {} or {} but got {}.'.format(
                             (n_qubits, n_qubits), (n_qubits, 2 * n_qubits),
                             shape))

    if isinstance(initial_state, int):
        initial_state = _occupied_orbitals(initial_state, n_qubits)
    initially_occupied_orbitals = cast(Optional[Sequence[int]], initial_state)

    # If the transformation matrix is block diagonal with two blocks,
    # do each block separately
    if _is_spin_block_diagonal(transformation_matrix):
        up_block = transformation_matrix[:n_qubits // 2, :n_qubits // 2]
        up_qubits = qubits[:n_qubits // 2]

        down_block = transformation_matrix[n_qubits // 2:, n_qubits // 2:]
        down_qubits = qubits[n_qubits // 2:]

        if initially_occupied_orbitals is None:
            up_orbitals = None
            down_orbitals = None
        else:
            up_orbitals = [
                i for i in initially_occupied_orbitals if i < n_qubits // 2
            ]
            down_orbitals = [
                i - n_qubits // 2
                for i in initially_occupied_orbitals
                if i >= n_qubits // 2
            ]

        yield bogoliubov_transform(up_qubits,
                                   up_block,
                                   initial_state=up_orbitals)
        yield bogoliubov_transform(down_qubits,
                                   down_block,
                                   initial_state=down_orbitals)
        return

    if shape == (n_qubits, n_qubits):
        # We're performing a particle-number conserving "Slater" basis change
        yield _slater_basis_change(qubits, transformation_matrix,
                                   initially_occupied_orbitals)
    else:
        # We're performing a more general Gaussian unitary
        yield _gaussian_basis_change(qubits, transformation_matrix,
                                     initially_occupied_orbitals)


def _is_spin_block_diagonal(matrix) -> bool:
    n = matrix.shape[0]
    if n % 2:
        return False
    max_upper_right = numpy.max(numpy.abs(matrix[:n // 2, n // 2:]))
    max_lower_left = numpy.max(numpy.abs(matrix[n // 2:, :n // 2]))
    return (numpy.isclose(max_upper_right, 0.0) and
            numpy.isclose(max_lower_left, 0.0))


def _occupied_orbitals(computational_basis_state: int, n_qubits) -> List[int]:
    """Indices of ones in the binary expansion of an integer in big endian
    order. e.g. 010110 -> [1, 3, 4]"""
    bitstring = format(computational_basis_state, 'b').zfill(n_qubits)
    return [j for j in range(len(bitstring)) if bitstring[j] == '1']


def _slater_basis_change(qubits: Sequence[cirq.Qid],
                         transformation_matrix: numpy.ndarray,
                         initially_occupied_orbitals: Optional[Sequence[int]]
                        ) -> cirq.OP_TREE:
    n_qubits = len(qubits)

    if initially_occupied_orbitals is None:
        decomposition, diagonal = linalg.givens_decomposition_square(
            transformation_matrix)
        circuit_description = list(reversed(decomposition))
        # The initial state is not a computational basis state so the
        # phases left on the diagonal in the decomposition matter
        yield (cirq.rz(rads=numpy.angle(diagonal[j])).on(qubits[j])
               for j in range(n_qubits))
    else:
        initially_occupied_orbitals = cast(Sequence[int],
                                           initially_occupied_orbitals)
        transformation_matrix = transformation_matrix[list(
            initially_occupied_orbitals)]
        n_occupied = len(initially_occupied_orbitals)
        # Flip bits so that the first n_occupied are 1 and the rest 0
        initially_occupied_orbitals_set = set(initially_occupied_orbitals)
        yield (cirq.X(qubits[j])
               for j in range(n_qubits)
               if (j < n_occupied) != (j in initially_occupied_orbitals_set))
        circuit_description = circuits.slater_determinant_preparation_circuit(
            transformation_matrix)

    yield _ops_from_givens_rotations_circuit_description(
        qubits, circuit_description)


def _gaussian_basis_change(qubits: Sequence[cirq.Qid],
                           transformation_matrix: numpy.ndarray,
                           initially_occupied_orbitals: Optional[Sequence[int]]
                          ) -> cirq.OP_TREE:
    n_qubits = len(qubits)

    # Rearrange the transformation matrix because the OpenFermion routine
    # expects it to describe annihilation operators rather than creation
    # operators
    left_block = transformation_matrix[:, :n_qubits]
    right_block = transformation_matrix[:, n_qubits:]
    transformation_matrix = numpy.block(
        [numpy.conjugate(right_block),
         numpy.conjugate(left_block)])

    decomposition, left_decomposition, _, left_diagonal = (
        linalg.fermionic_gaussian_decomposition(transformation_matrix))

    if (initially_occupied_orbitals is not None and
            len(initially_occupied_orbitals) == 0):
        # Starting with the vacuum state yields additional symmetry
        circuit_description = list(reversed(decomposition))
    else:
        if initially_occupied_orbitals is None:
            # The initial state is not a computational basis state so the
            # phases left on the diagonal in the Givens decomposition matter
            yield (cirq.rz(rads=numpy.angle(left_diagonal[j])).on(qubits[j])
                   for j in range(n_qubits))
        circuit_description = list(reversed(decomposition + left_decomposition))

    yield _ops_from_givens_rotations_circuit_description(
        qubits, circuit_description)


def _ops_from_givens_rotations_circuit_description(
        qubits: Sequence[cirq.Qid], circuit_description: Iterable[Iterable[
            Union[str, Tuple[int, int, float, float]]]]) -> cirq.OP_TREE:
    """Yield operations from a Givens rotations circuit obtained from
    OpenFermion.
    """
    for parallel_ops in circuit_description:
        for op in parallel_ops:
            if op == 'pht':
                yield cirq.X(qubits[-1])
            else:
                i, j, theta, phi = cast(Tuple[int, int, float, float], op)
                yield circuits.Ryxxy(theta).on(qubits[i], qubits[j])
                yield cirq.Z(qubits[j])**(phi / numpy.pi)
